@@ -3,6 +3,7 @@
 OCKCONFDIR="$1"
 EPCONFDIR="$2"
 CCACONFDIR="$3"
+COMBINED_EXTRACT_FILE="$4"
 
 LATESTCEXP="CEX8P"
 
@@ -27,7 +28,15 @@ EOF
     if [ "x$4" != "x" ]; then
        echo "confname = $4" >> "${OCKCONFDIR}/opencryptoki.conf"
     fi
+    if test ! -z ${TOKEN_GROUP}; then
+       echo "usergroup = ${TOKEN_GROUP}" >> "${OCKCONFDIR}/opencryptoki.conf"
+    fi
     echo "}" >> "${OCKCONFDIR}/opencryptoki.conf"
+
+    pkcstok_admin remove --token $3 --force &> /dev/null
+    if test ! -z ${TOKEN_GROUP}; then
+        pkcstok_admin create --token $3 --group ${TOKEN_GROUP} --force &> /dev/null
+    fi
 }
 
 # Usage: genep11cfg num configline
@@ -65,18 +74,29 @@ EXPECTED_MKVPS {
   APKA = "$CCA_APKA_MKVP"
 }
 PKEY_MODE = ENABLED
+AES_KEY_MODE = CIPHER
 EOF
 }
 
-if test $(($(date +%j)%2)) == 1; then
+if test $(($(date +%-j)%2)) == 1; then
     USENEWFORMAT=/bin/true
     echo "Using FIPS compliant token store"
 else
     echo "Using legacy token store"
 fi
 
+if test ! -z ${PKCS11_TEST_USER}; then
+    if test ! -z ${PKCS11_TEST_GROUP}; then
+        TOKEN_GROUP=${PKCS11_TEST_GROUP}
+    else
+        TOKEN_GROUP="tokgroup"
+    fi
+    getent group ${TOKEN_GROUP} >/dev/null || groupadd -r ${TOKEN_GROUP}
+    usermod -a -G ${TOKEN_GROUP} ${PKCS11_TEST_USER}
+fi
+
 # initialize opencryptoki.conf
-echo "version opencryptoki-3.23" > "${OCKCONFDIR}/opencryptoki.conf"
+echo "version opencryptoki-3.25" > "${OCKCONFDIR}/opencryptoki.conf"
 
 # enable full statistics
 echo "statistics (on,implicit,internal)" >> "${OCKCONFDIR}/opencryptoki.conf"
@@ -132,3 +152,11 @@ addslot 44 libpkcs11_ep11.so ep4 ep11tok44.conf
 if genlatestep11cfg 45 "PKEY_MODE ENABLE4NONEXTR"; then
     addslot 45 libpkcs11_ep11.so ep5 ep11tok45.conf
 fi
+
+# 6: latest (CEX8 only)
+# PKEY_MODE ENABLE4ALL
+if genlatestep11cfg 46 "PKEY_MODE ENABLE4ALL"; then
+    addslot 46 libpkcs11_ep11.so ep6 ep11tok46.conf
+    echo "46" > $COMBINED_EXTRACT_FILE
+fi
+
