@@ -15,6 +15,7 @@
 
 #include "pkcs11types.h"
 #include "regress.h"
+#include "platform.h"
 
 #define UNUSED(var)            ((void)(var))
 
@@ -35,6 +36,7 @@ CK_BBOOL securekey;
  * pkey option.
  */
 CK_BBOOL pkey = CK_FALSE;
+CK_BBOOL combined_extract = CK_FALSE;
 
 CK_ULONG t_total = 0;           // total test assertions
 CK_ULONG t_ran = 0;             // number of assertions ran
@@ -53,7 +55,9 @@ static void *pkcs11lib = NULL;
 static void unload_pkcslib(void)
 {
     if (pkcs11lib != NULL) {
+#ifndef WITH_SANITIZER
          dlclose(pkcs11lib);
+#endif
     }
 }
 
@@ -161,6 +165,11 @@ int is_rejected_by_policy(CK_RV ret_code, CK_SESSION_HANDLE session)
         return 0;
     }
 
+    if (info.ulDeviceError != 0 && info.ulDeviceError != CKR_POLICY_VIOLATION) {
+        testcase_notice("CK_SESSION_INFO.ulDeviceError: 0x%lx",
+                        info.ulDeviceError);
+    }
+
     return (info.ulDeviceError == CKR_POLICY_VIOLATION);
 }
 
@@ -185,6 +194,9 @@ CK_RV create_AESKey(CK_SESSION_HANDLE session, CK_BBOOL extractable,
     };
     CK_ULONG keyTemplate_len = sizeof(keyTemplate) / sizeof(CK_ATTRIBUTE);
 
+    if (combined_extract)
+        pkeyextractable = CK_TRUE;
+
     rc = funcs->C_CreateObject(session, keyTemplate, keyTemplate_len, h_key);
     if (rc != CKR_OK) {
         if (is_rejected_by_policy(rc, session))
@@ -209,6 +221,9 @@ CK_RV generate_AESKey(CK_SESSION_HANDLE session,
     };
     CK_ULONG key_gen_tmpl_len = sizeof(key_gen_tmpl) / sizeof(CK_ATTRIBUTE);
 
+    if (combined_extract)
+        pkeyextractable = CK_TRUE;
+
     CK_RV rc = funcs->C_GenerateKey(session, mechkey,
                                     key_gen_tmpl, key_gen_tmpl_len,
                                     h_key);
@@ -395,9 +410,17 @@ CK_RV create_RSAPrivateKey(CK_SESSION_HANDLE session,
         {CKA_EXPONENT_2, exponent2, exponent2_len},
         {CKA_COEFFICIENT, coefficient, coefficient_len}
     };
+    CK_ULONG template_len = sizeof(template) / sizeof(CK_ATTRIBUTE);
+
+    if (prime1 == NULL || prime1_len == 0 ||
+        prime2 == NULL || prime2_len == 0 ||
+        exponent1 == NULL || exponent1_len == 0 ||
+        exponent2 == NULL || exponent2_len == 0 ||
+        coefficient == NULL || coefficient_len == 0)
+        template_len -= 5;
 
     // create key
-    rc = funcs->C_CreateObject(session, template, 17, priv_key);
+    rc = funcs->C_CreateObject(session, template, template_len, priv_key);
     if (rc != CKR_OK) {
         if (is_rejected_by_policy(rc, session))
             rc = CKR_POLICY_VIOLATION;
@@ -492,6 +515,107 @@ CK_RV generate_RSA_PKCS_KeyPair(CK_SESSION_HANDLE session,
     // see rsa_func.c
 }
 
+struct rsa_key_cache_entry {
+    CK_SESSION_HANDLE session;
+    CK_ULONG modulusBits;
+    CK_BYTE *publicExponent;
+    CK_ULONG publicExponent_len;
+    CK_OBJECT_HANDLE publ_key;
+    CK_OBJECT_HANDLE priv_key;
+};
+
+struct rsa_key_cache_entry *rsa_key_cache = NULL;
+CK_ULONG rsa_key_cache_size = 0;
+
+CK_RV generate_RSA_PKCS_KeyPair_cached(CK_SESSION_HANDLE session,
+                                       CK_ULONG modulusBits,
+                                       CK_BYTE publicExponent[],
+                                       CK_ULONG publicExponent_len,
+                                       CK_OBJECT_HANDLE *publ_key,
+                                       CK_OBJECT_HANDLE *priv_key)
+{
+    struct rsa_key_cache_entry *tmp, *free = NULL;
+    CK_ULONG i;
+    CK_RV rc;
+
+    for (i = 0; i < rsa_key_cache_size; i++) {
+        if (rsa_key_cache[i].session == session &&
+            rsa_key_cache[i].modulusBits == modulusBits &&
+            rsa_key_cache[i].publicExponent_len == publicExponent_len &&
+            memcmp(rsa_key_cache[i].publicExponent, publicExponent,
+                   publicExponent_len) == 0) {
+            *publ_key = rsa_key_cache[i].publ_key;
+            *priv_key = rsa_key_cache[i].priv_key;
+
+            return CKR_OK;
+        }
+
+        if (rsa_key_cache[i].session == CK_INVALID_HANDLE && free == NULL)
+            free = &rsa_key_cache[i];
+    }
+
+    rc = generate_RSA_PKCS_KeyPair(session, modulusBits, publicExponent,
+                                   publicExponent_len, publ_key, priv_key);
+    if (rc != CKR_OK)
+        return rc;
+
+    if (free == NULL) {
+        tmp = realloc(rsa_key_cache, (rsa_key_cache_size + 1) *
+                                            sizeof(struct rsa_key_cache_entry));
+        if (tmp == NULL) {
+            testcase_error("realloc failed to enlarge the RSA key cache");
+            return CKR_HOST_MEMORY;
+        }
+
+        free = &tmp[rsa_key_cache_size];
+        memset(free, 0, sizeof(*free));
+
+        rsa_key_cache = tmp;
+        rsa_key_cache_size++;
+    }
+
+    free->session = session;
+    free->modulusBits = modulusBits;
+    free->publicExponent_len = publicExponent_len;
+    free->publicExponent = malloc(publicExponent_len);
+    if (free->publicExponent == NULL) {
+        testcase_error("failed to allocate the public exponent cache entry");
+        return CKR_HOST_MEMORY;
+    }
+    memcpy(free->publicExponent, publicExponent, publicExponent_len);
+    free->publ_key = *publ_key;
+    free->priv_key = *priv_key;
+
+    return CKR_OK;
+}
+
+void free_rsa_key_cache(CK_SESSION_HANDLE session)
+{
+    CK_ULONG i;
+
+    for (i = 0; i < rsa_key_cache_size; i++) {
+        if (rsa_key_cache[i].session == CK_INVALID_HANDLE)
+            continue;
+        if (session != CK_INVALID_HANDLE && rsa_key_cache[i].session != session)
+            continue;
+
+        funcs->C_DestroyObject(rsa_key_cache[i].session,
+                               rsa_key_cache[i].publ_key);
+        funcs->C_DestroyObject(rsa_key_cache[i].session,
+                               rsa_key_cache[i].priv_key);
+        free(rsa_key_cache[i].publicExponent);
+
+        /* mark as free */
+        memset(&rsa_key_cache[i], 0, sizeof(rsa_key_cache[i]));
+    }
+
+    if (session == CK_INVALID_HANDLE) {
+        free(rsa_key_cache);
+        rsa_key_cache = NULL;
+        rsa_key_cache_size = 0;
+    }
+}
+
 /** Generate an EC key pair **/
 CK_RV generate_EC_KeyPair(CK_SESSION_HANDLE session,
                           CK_BYTE* ec_params, CK_ULONG ec_params_len,
@@ -508,6 +632,7 @@ CK_RV generate_EC_KeyPair(CK_SESSION_HANDLE session,
     CK_ATTRIBUTE publicKeyTemplate[] = {
         {CKA_VERIFY, &true, sizeof(true)},
         {CKA_EC_PARAMS, ec_params, ec_params_len},
+        {CKA_IBM_PROTKEY_EXTRACTABLE, &pkeyextractable, sizeof(CK_BBOOL)},
     };
     CK_ATTRIBUTE privateKeyTemplate[] = {
         {CKA_TOKEN, &true, sizeof(true)},
@@ -520,8 +645,11 @@ CK_RV generate_EC_KeyPair(CK_SESSION_HANDLE session,
         {CKA_EXTRACTABLE, &extractable, sizeof(CK_BBOOL)},
         {CKA_IBM_PROTKEY_EXTRACTABLE, &pkeyextractable, sizeof(CK_BBOOL)},
     };
-    CK_ULONG num_publ_attrs = sizeof(publicKeyTemplate)/sizeof(CK_ATTRIBUTE);
-    CK_ULONG num_priv_attrs = sizeof(privateKeyTemplate)/sizeof(CK_ATTRIBUTE);
+    CK_ULONG num_publ_attrs = sizeof(publicKeyTemplate) / sizeof(CK_ATTRIBUTE);
+    CK_ULONG num_priv_attrs = sizeof(privateKeyTemplate) / sizeof(CK_ATTRIBUTE);
+
+    if (combined_extract)
+        pkeyextractable = CK_TRUE;
 
     // generate keys
     rc = funcs->C_GenerateKeyPair(session,
@@ -544,7 +672,8 @@ CK_RV create_ECPrivateKey(CK_SESSION_HANDLE session,
                           CK_BYTE privatekey[],
                           CK_ULONG privatekey_len,
                           CK_OBJECT_HANDLE * priv_key,
-                          CK_BBOOL extractable)
+                          CK_BBOOL extractable,
+                          CK_BBOOL derive)
 {
 
     CK_OBJECT_CLASS class = CKO_PRIVATE_KEY;
@@ -566,13 +695,16 @@ CK_RV create_ECPrivateKey(CK_SESSION_HANDLE session,
         {CKA_ID, id, sizeof(id)},
         {CKA_SENSITIVE, &true, sizeof(true)},
         {CKA_SIGN, &true, sizeof(true)},
-        {CKA_DERIVE, &true, sizeof(true)},
+        {CKA_DERIVE, &derive, sizeof(derive)},
         {CKA_EC_PARAMS, params, params_len},
         {CKA_VALUE, privatekey, privatekey_len},
         {CKA_EXTRACTABLE, &extractable, sizeof(CK_BBOOL)},
         {CKA_IBM_PROTKEY_EXTRACTABLE, &pkeyextractable, sizeof(CK_BBOOL)},
     };
 
+    if (combined_extract)
+        pkeyextractable = CK_TRUE;
+
     // create key
     rc = funcs->C_CreateObject(session, template,
                                sizeof(template) / sizeof(CK_ATTRIBUTE),
@@ -589,10 +721,12 @@ CK_RV create_ECPublicKey(CK_SESSION_HANDLE session,
                          CK_BYTE params[],
                          CK_ULONG params_len,
                          CK_BYTE pointq[],
-                         CK_ULONG pointq_len, CK_OBJECT_HANDLE * publ_key)
+                         CK_ULONG pointq_len, CK_OBJECT_HANDLE * publ_key,
+                         CK_BBOOL extractable)
 {
     CK_RV rc;
     CK_OBJECT_CLASS class = CKO_PUBLIC_KEY;
+    CK_BBOOL pkeyextractable = !extractable;
     CK_KEY_TYPE keyType = CKK_EC;
     CK_UTF8CHAR label[] = "An EC public key object";
     CK_BBOOL true = TRUE;
@@ -604,7 +738,8 @@ CK_RV create_ECPublicKey(CK_SESSION_HANDLE session,
         {CKA_VERIFY, &true, sizeof(true)},
         {CKA_DERIVE, &true, sizeof(true)},
         {CKA_EC_PARAMS, params, params_len},
-        {CKA_EC_POINT, pointq, pointq_len}
+        {CKA_EC_POINT, pointq, pointq_len},
+        {CKA_IBM_PROTKEY_EXTRACTABLE, &pkeyextractable, sizeof(CK_BBOOL)},
     };
 
     // create key
@@ -618,6 +753,53 @@ CK_RV create_ECPublicKey(CK_SESSION_HANDLE session,
     return rc;
 }
 
+/** Generate an EC key pair **/
+CK_RV generate_Dilithium_KeyPair(CK_SESSION_HANDLE session,
+                                 CK_ULONG keyform,
+                                 CK_OBJECT_HANDLE *publ_key,
+                                 CK_OBJECT_HANDLE *priv_key,
+                                 CK_BBOOL extractable)
+{
+    CK_RV rc;
+    CK_MECHANISM mech = { CKM_IBM_DILITHIUM, NULL, 0 };
+    CK_BBOOL pkeyextractable = !extractable;
+    CK_BYTE subject[] = {0};
+    CK_BYTE id[] = { 123 };
+    CK_BBOOL true = TRUE;
+    CK_ATTRIBUTE publicKeyTemplate[] = {
+        {CKA_VERIFY, &true, sizeof(true)},
+        {CKA_IBM_DILITHIUM_KEYFORM, &keyform, sizeof(keyform)},
+    };
+    CK_ATTRIBUTE privateKeyTemplate[] = {
+        {CKA_TOKEN, &true, sizeof(true)},
+        {CKA_PRIVATE, &true, sizeof(true)},
+        {CKA_SUBJECT, subject, 0},
+        {CKA_ID, id, sizeof(id)},
+        {CKA_SENSITIVE, &true, sizeof(true)},
+        {CKA_SIGN, &true, sizeof(true)},
+        {CKA_EXTRACTABLE, &extractable, sizeof(CK_BBOOL)},
+        {CKA_IBM_PROTKEY_EXTRACTABLE, &pkeyextractable, sizeof(CK_BBOOL)},
+        {CKA_IBM_DILITHIUM_KEYFORM, &keyform, sizeof(keyform)},
+    };
+    CK_ULONG num_publ_attrs = sizeof(publicKeyTemplate) / sizeof(CK_ATTRIBUTE);
+    CK_ULONG num_priv_attrs = sizeof(privateKeyTemplate) / sizeof(CK_ATTRIBUTE);
+
+    if (combined_extract)
+        pkeyextractable = CK_TRUE;
+
+    // generate keys
+    rc = funcs->C_GenerateKeyPair(session,
+                                  &mech,
+                                  publicKeyTemplate, num_publ_attrs,
+                                  privateKeyTemplate, num_priv_attrs,
+                                  publ_key, priv_key);
+
+    if (is_rejected_by_policy(rc, session))
+        rc = CKR_POLICY_VIOLATION;
+
+    return rc;
+}
+
 /** Create an IBM Dilithium private key using private values **/
 CK_RV create_DilithiumPrivateKey(CK_SESSION_HANDLE session,
                                  CK_BYTE pkcs8[], CK_ULONG pkcs8_len,
@@ -1361,6 +1543,8 @@ int do_ParseArgs(int argc, char **argv)
             no_init = TRUE;
         } else if (strcmp(argv[i], "-nostop") == 0) {
             no_stop = TRUE;
+        } else if (strcmp(argv[i], "-combined-extract") == 0) {
+            combined_extract = TRUE;
         } else {
             printf("Invalid argument passed as option: %s\n", argv[i]);
             usage(argv[0]);
@@ -1392,7 +1576,7 @@ CK_BBOOL do_GetFunctionList(void)
     CK_RV(*getinterface)(CK_UTF8CHAR_PTR name, CK_VERSION_PTR version,
                          CK_INTERFACE_PTR_PTR interface, CK_FLAGS flags);
     char *e;
-    char *f = "libopencryptoki.so";
+    char *f = OCK_API_LIBNAME;
     CK_ULONG nmemb = 0;
 
     rv = FALSE;
@@ -1401,7 +1585,7 @@ CK_BBOOL do_GetFunctionList(void)
     if (e == NULL)
         e = f;
 
-    pkcs11lib = dlopen(e, RTLD_NOW);
+    pkcs11lib = dlopen(e, DYNLIB_LDFLAGS);
     if (pkcs11lib == NULL)
         goto ret;
 
@@ -1458,7 +1642,9 @@ ret:
         ifs = NULL;
 
         if (pkcs11lib != NULL) {
+#ifndef WITH_SANITIZER
             dlclose(pkcs11lib);
+#endif
             pkcs11lib = NULL;
 	}
     }
