@@ -57,30 +57,31 @@ typedef struct kemParam {
     CK_ULONG shard_data_len;
     CK_BYTE shared_data[32];
     CK_BBOOL hybrid;
+    CK_BBOOL prepend;
 } _kemParam;
 
 const _kemParam kemInput[] = {
-    { CKD_NULL, 32, 0, {0x00}, CK_FALSE },
-    { CKD_IBM_HYBRID_SHA1_KDF, 16, 0, {0x00}, CK_TRUE },
+    { CKD_NULL, 32, 0, {0x00}, CK_FALSE, CK_FALSE },
+    { CKD_IBM_HYBRID_SHA1_KDF, 16, 0, {0x00}, CK_TRUE, CK_FALSE },
     { CKD_IBM_HYBRID_SHA1_KDF, 32, 16,
       { 0x32,0x3F,0xA3,0x16,0x9D,0x8E,0x9C,0x65,0x93,0xF5,0x94,0x76,0xBC,0x14,0x20,0x00 },
-      CK_TRUE },
-    { CKD_IBM_HYBRID_SHA224_KDF, 16, 0, {0x00}, CK_TRUE },
+      CK_TRUE, CK_TRUE },
+    { CKD_IBM_HYBRID_SHA224_KDF, 16, 0, {0x00}, CK_TRUE, CK_FALSE },
     { CKD_IBM_HYBRID_SHA224_KDF, 32, 16,
       { 0x32,0x3F,0xA3,0x16,0x9D,0x8E,0x9C,0x65,0x93,0xF5,0x94,0x76,0xBC,0x14,0x20,0x00 },
-      CK_TRUE },
-    { CKD_IBM_HYBRID_SHA256_KDF, 16, 0, {0x00}, CK_TRUE },
+      CK_TRUE, CK_TRUE },
+    { CKD_IBM_HYBRID_SHA256_KDF, 16, 0, {0x00}, CK_TRUE, CK_FALSE },
     { CKD_IBM_HYBRID_SHA256_KDF, 32, 16,
       { 0x32,0x3F,0xA3,0x16,0x9D,0x8E,0x9C,0x65,0x93,0xF5,0x94,0x76,0xBC,0x14,0x20,0x00 },
-      CK_TRUE },
-    { CKD_IBM_HYBRID_SHA384_KDF, 16, 0, {0x00}, CK_TRUE },
+      CK_TRUE, CK_TRUE },
+    { CKD_IBM_HYBRID_SHA384_KDF, 16, 0, {0x00}, CK_TRUE, CK_FALSE },
     { CKD_IBM_HYBRID_SHA384_KDF, 32, 16,
       { 0x32,0x3F,0xA3,0x16,0x9D,0x8E,0x9C,0x65,0x93,0xF5,0x94,0x76,0xBC,0x14,0x20,0x00 },
-      CK_TRUE },
-    { CKD_IBM_HYBRID_SHA512_KDF, 16, 0, {0x00}, CK_TRUE },
+      CK_TRUE, CK_TRUE },
+    { CKD_IBM_HYBRID_SHA512_KDF, 16, 0, {0x00}, CK_TRUE, CK_FALSE },
     { CKD_IBM_HYBRID_SHA512_KDF, 32, 16,
       { 0x32,0x3F,0xA3,0x16,0x9D,0x8E,0x9C,0x65,0x93,0xF5,0x94,0x76,0xBC,0x14,0x20,0x00 },
-      CK_TRUE },
+      CK_TRUE, CK_TRUE },
 };
 
 static const char *p11_get_ckd(CK_EC_KDF_TYPE kdf)
@@ -250,6 +251,7 @@ CK_RV run_EnDecapsulateKyber(CK_SESSION_HANDLE session,
                              CK_ULONG secret_key_len,
                              CK_IBM_KYBER_KDF_TYPE kdf,
                              CK_BBOOL hybrid,
+                             CK_BBOOL prepend,
                              const CK_BYTE *pSharedData,
                              CK_ULONG ulSharedDataLen,
                              const CK_BYTE *pCipher,
@@ -258,7 +260,6 @@ CK_RV run_EnDecapsulateKyber(CK_SESSION_HANDLE session,
                              CK_ULONG ulExpectedSecret)
 {
     CK_MECHANISM mech;
-    CK_MECHANISM_INFO mech_info;
     CK_IBM_KYBER_PARAMS kyber_params;
     CK_BYTE *cipher = NULL;
     CK_ULONG cipher_len = 0;
@@ -287,19 +288,12 @@ CK_RV run_EnDecapsulateKyber(CK_SESSION_HANDLE session,
     mech.pParameter = &kyber_params;
 
     /* Query the slot, check if this mech if supported */
-    rc = funcs->C_GetMechanismInfo(SLOT_ID, mech.mechanism, &mech_info);
-    if (rc != CKR_OK) {
-        if (rc == CKR_MECHANISM_INVALID) {
-            /* no support for Kyber? skip */
-            testcase_skip("Slot %u doesn't support %s",
-                          (unsigned int) SLOT_ID,
-                          p11_get_ckm(&mechtable_funcs, mech.mechanism));
-            rc = CKR_OK;
-            goto testcase_cleanup;
-        } else {
-            testcase_error("C_GetMechanismInfo() rc = %s", p11_get_ckr(rc));
-            goto testcase_cleanup;
-        }
+    if (!mech_supported(SLOT_ID, mech.mechanism)) {
+        testcase_skip("Slot %u doesn't support %s",
+                      (unsigned int) SLOT_ID,
+                      p11_get_ckm(&mechtable_funcs, mech.mechanism));
+        rc = CKR_OK;
+        goto testcase_cleanup;
     }
 
     if (hybrid) {
@@ -323,7 +317,7 @@ CK_RV run_EnDecapsulateKyber(CK_SESSION_HANDLE session,
     kyber_params.kdf = kdf;
     kyber_params.pSharedData = (CK_BYTE *)pSharedData;
     kyber_params.ulSharedDataLen = ulSharedDataLen;
-    kyber_params.bPrepend = (hybrid_key != CK_INVALID_HANDLE);
+    kyber_params.bPrepend = prepend;
     kyber_params.hSecret = hybrid_key;
 
     /* Size query */
@@ -363,7 +357,7 @@ decapsulate:
     kyber_params.kdf = kdf;
     kyber_params.pSharedData = (CK_BYTE *)pSharedData;
     kyber_params.ulSharedDataLen = ulSharedDataLen;
-    kyber_params.bPrepend = (hybrid_key != CK_INVALID_HANDLE);
+    kyber_params.bPrepend = prepend;
     kyber_params.hSecret = hybrid_key;
     kyber_params.ulCipherLen = cipher_len;
     kyber_params.pCipher = cipher;
@@ -426,7 +420,6 @@ CK_RV run_EncrypDecryptKyber(CK_SESSION_HANDLE session,
     CK_MECHANISM mech;
     CK_BYTE_PTR data = NULL, encrypted = NULL, decrypted = NULL;
     CK_ULONG i, datalen, encrypted_len, decrypted_len;
-    CK_MECHANISM_INFO mech_info;
     CK_RV rc;
 
     mech.mechanism = CKM_IBM_KYBER;
@@ -436,19 +429,12 @@ CK_RV run_EncrypDecryptKyber(CK_SESSION_HANDLE session,
     datalen = 32; /* Kyber can encrypt blocks of 32 bytes */
 
     /* Query the slot, check if this mech if supported */
-    rc = funcs->C_GetMechanismInfo(SLOT_ID, mech.mechanism, &mech_info);
-    if (rc != CKR_OK) {
-        if (rc == CKR_MECHANISM_INVALID) {
-            /* no support for Kyber? skip */
-            testcase_skip("Slot %u doesn't support %s",
-                          (unsigned int) SLOT_ID,
-                          p11_get_ckm(&mechtable_funcs, mech.mechanism));
-            rc = CKR_OK;
-            goto testcase_cleanup;
-        } else {
-            testcase_error("C_GetMechanismInfo() rc = %s", p11_get_ckr(rc));
-            goto testcase_cleanup;
-        }
+    if (!mech_supported(SLOT_ID, mech.mechanism)) {
+        testcase_skip("Slot %u doesn't support %s",
+                      (unsigned int) SLOT_ID,
+                      p11_get_ckm(&mechtable_funcs, mech.mechanism));
+        rc = CKR_OK;
+        goto testcase_cleanup;
     }
 
     data = calloc(datalen, sizeof(CK_BYTE));
@@ -547,7 +533,6 @@ CK_RV run_GenerateKyberKeyPairEnDecryptKEM(void)
     CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
     CK_ULONG user_pin_len, i, j;
     CK_FLAGS flags;
-    CK_MECHANISM_INFO mech_info;
     CK_RV rc;
 
     testcase_rw_session();
@@ -558,18 +543,12 @@ CK_RV run_GenerateKyberKeyPairEnDecryptKEM(void)
     mech.pParameter = NULL;
 
     /* query the slot, check if this mech is supported */
-    rc = funcs->C_GetMechanismInfo(SLOT_ID, mech.mechanism, &mech_info);
-    if (rc != CKR_OK) {
-        if (rc == CKR_MECHANISM_INVALID) {
-            /* no support for Kyber key gen? skip */
-            testcase_skip("Slot %u doesn't support CKM_IBM_KYBER ",
-                          (unsigned int) SLOT_ID);
-            rc = CKR_OK;
-            goto testcase_cleanup;
-        } else {
-            testcase_error("C_GetMechanismInfo() rc = %s", p11_get_ckr(rc));
-            goto testcase_cleanup;
-        }
+    if (!mech_supported(SLOT_ID, mech.mechanism)) {
+        testcase_skip("Slot %u doesn't support %s",
+                      (unsigned int) SLOT_ID,
+                      p11_get_ckm(&mechtable_funcs, mech.mechanism));
+        rc = CKR_OK;
+        goto testcase_cleanup;
     }
 
     for (i = 0; i < 2 * num_variants; i++) {
@@ -653,6 +632,7 @@ CK_RV run_GenerateKyberKeyPairEnDecryptKEM(void)
                                         kemInput[j].secret_key_len,
                                         kemInput[j].kdf,
                                         kemInput[j].hybrid,
+                                        kemInput[j].prepend,
                                         kemInput[j].shared_data,
                                         kemInput[j].shard_data_len,
                                         NULL, 0, NULL, 0);
@@ -697,34 +677,22 @@ testcase_cleanup:
 
 CK_RV run_ImportKyberKeyPairKEM(void)
 {
-    CK_MECHANISM mech;
     CK_OBJECT_HANDLE publ_key = CK_INVALID_HANDLE, priv_key = CK_INVALID_HANDLE;
     CK_SESSION_HANDLE session;
     CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
     CK_ULONG user_pin_len, i;
     CK_FLAGS flags;
-    CK_MECHANISM_INFO mech_info;
     CK_RV rc;
 
     testcase_rw_session();
     testcase_user_login();
 
-    mech.mechanism = CKM_IBM_KYBER;
-    mech.ulParameterLen = 0;
-    mech.pParameter = NULL;
-
     /* query the slot, check if this mech is supported */
-    rc = funcs->C_GetMechanismInfo(SLOT_ID, mech.mechanism, &mech_info);
-    if (rc != CKR_OK) {
-        if (rc == CKR_MECHANISM_INVALID) {
-            /* no support for Kyber key gen? skip */
-            testcase_skip("Slot %u doesn't support CKM_IBM_KYBER",
-                          (unsigned int) SLOT_ID);
-            goto testcase_cleanup;
-        } else {
-            testcase_error("C_GetMechanismInfo() rc = %s", p11_get_ckr(rc));
-            goto testcase_cleanup;
-        }
+    if (!mech_supported(SLOT_ID, CKM_IBM_KYBER)) {
+        testcase_skip("Slot %u doesn't support CKM_IBM_KYBER",
+                      (unsigned int) SLOT_ID);
+        rc = CKR_OK;
+        goto testcase_cleanup;
     }
 
     for (i = 0; i < KYBER_TV_NUM; i++) {
@@ -786,7 +754,7 @@ CK_RV run_ImportKyberKeyPairKEM(void)
         testcase_new_assertion();
         rc = run_EnDecapsulateKyber(session, priv_key, publ_key,
                                     kyber_tv[i].secret_len,
-                                    CKD_NULL, CK_FALSE, NULL, 0,
+                                    CKD_NULL, CK_FALSE, CK_FALSE, NULL, 0,
                                     kyber_tv[i].cipher, kyber_tv[i].cipher_len,
                                     kyber_tv[i].secret, kyber_tv[i].secret_len);
         if (rc != 0) {
@@ -918,13 +886,11 @@ done:
 
 CK_RV run_TransferKyberKeyPair(void)
 {
-    CK_MECHANISM mech;
     CK_OBJECT_HANDLE publ_key = CK_INVALID_HANDLE, priv_key = CK_INVALID_HANDLE;
     CK_SESSION_HANDLE session;
     CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
     CK_ULONG user_pin_len, i;
     CK_FLAGS flags;
-    CK_MECHANISM_INFO mech_info;
     CK_RV rc;
     CK_OBJECT_HANDLE secret_key = CK_INVALID_HANDLE;
     CK_BYTE_PTR wrapped_key = NULL;
@@ -935,22 +901,24 @@ CK_RV run_TransferKyberKeyPair(void)
     testcase_rw_session();
     testcase_user_login();
 
-    mech.mechanism = CKM_IBM_KYBER;
-    mech.ulParameterLen = 0;
-    mech.pParameter = NULL;
-
     /* query the slot, check if this mech is supported */
-    rc = funcs->C_GetMechanismInfo(SLOT_ID, mech.mechanism, &mech_info);
-    if (rc != CKR_OK) {
-        if (rc == CKR_MECHANISM_INVALID) {
-            /* no support for Kyber key gen? skip */
-            testcase_skip("Slot %u doesn't support CKM_IBM_KYBER",
-                          (unsigned int) SLOT_ID);
-            goto testcase_cleanup;
-        } else {
-            testcase_error("C_GetMechanismInfo() rc = %s", p11_get_ckr(rc));
-            goto testcase_cleanup;
-        }
+    if (!mech_supported(SLOT_ID, CKM_IBM_KYBER)) {
+        testcase_skip("Slot %u doesn't support CKM_IBM_KYBER",
+                      (unsigned int) SLOT_ID);
+        rc = CKR_OK;
+        goto testcase_cleanup;
+    }
+    if (!mech_supported(SLOT_ID, CKM_AES_KEY_GEN)) {
+        testcase_skip("Slot %u doesn't support CKM_AES_KEY_GEN",
+                      (unsigned int) SLOT_ID);
+        rc = CKR_OK;
+        goto testcase_cleanup;
+    }
+    if (!mech_supported_flags(SLOT_ID, CKM_AES_CBC_PAD, CKF_WRAP)) {
+        testcase_skip("Slot %u doesn't support key wrapping with CKM_AES_CBC_PAD",
+                      (unsigned int) SLOT_ID);
+        rc = CKR_OK;
+        goto testcase_cleanup;
     }
 
     for (i = 0; i < KYBER_TV_NUM; i++) {
@@ -1071,7 +1039,7 @@ CK_RV run_TransferKyberKeyPair(void)
         testcase_new_assertion();
         rc = run_EnDecapsulateKyber(session, unwrapped_key, publ_key,
                                     kyber_tv[i].secret_len,
-                                    CKD_NULL, CK_FALSE, NULL, 0,
+                                    CKD_NULL, CK_FALSE, CK_FALSE, NULL, 0,
                                     kyber_tv[i].cipher, kyber_tv[i].cipher_len,
                                     kyber_tv[i].secret, kyber_tv[i].secret_len);
         if (rc != 0) {
