@@ -61,11 +61,35 @@ CK_RSA_PKCS_OAEP_PARAMS oaep_params_sha256_source = {
         .ulSourceDataLen = 3,
 };
 
+CK_RSA_AES_KEY_WRAP_PARAMS rsa_aeskw_aes_256_oaep_sha1_params = {
+        .ulAESKeyBits = 256,
+        .pOAEPParams = &oaep_params_sha1,
+};
+
 CK_BYTE aes_iv[16] = { 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
                        0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f };
 
 CK_BYTE des_iv[8] = { 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07 };
 
+CK_ECDH_AES_KEY_WRAP_PARAMS ecdh_aeskw_aes_256__params = {
+        .ulAESKeyBits = 256,
+        .kdf = CKD_SHA256_KDF,
+        .ulSharedDataLen = 3,
+        .pSharedData = (CK_BYTE *)"abc",
+};
+
+CK_BYTE gcm_aad[8] = { 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07 };
+
+CK_GCM_PARAMS aes_gcm_param = {
+       .pIv = aes_iv,
+       .ulIvLen = sizeof(aes_iv),
+       .pAAD = gcm_aad,
+       .ulAADLen = sizeof(gcm_aad),
+       .ulTagBits = 32,
+};
+
+CK_BYTE prime256v1[] = OCK_PRIME256V1;
+
 struct wrapping_mech_info {
     char *name;
     CK_MECHANISM wrapping_mech;
@@ -74,6 +98,8 @@ struct wrapping_mech_info {
     CK_ULONG rsa_publ_exp_len;
     CK_BYTE rsa_publ_exp[4];
     CK_ULONG sym_keylen;
+    CK_BYTE ec_parms_len;
+    CK_BYTE *ec_parms;
 };
 
 struct wrapping_mech_info wrapping_tests[] = {
@@ -191,6 +217,24 @@ struct wrapping_mech_info wrapping_tests[] = {
         .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
         .sym_keylen = 32,
     },
+    {
+        .name = "Wrap/Unwrap with AES 128 GCM",
+        .wrapping_mech = { CKM_AES_GCM, &aes_gcm_param, sizeof(aes_gcm_param) },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 16,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 192 GCM",
+        .wrapping_mech = { CKM_AES_GCM, &aes_gcm_param, sizeof(aes_gcm_param) },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 24,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 GCM",
+        .wrapping_mech = { CKM_AES_GCM, &aes_gcm_param, sizeof(aes_gcm_param) },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
     {
         .name = "Wrap/Unwrap with AES 128 XTS",
         .wrapping_mech = { CKM_AES_XTS, aes_iv, sizeof(aes_iv) },
@@ -203,6 +247,107 @@ struct wrapping_mech_info wrapping_tests[] = {
         .wrapping_key_gen_mech = { CKM_AES_XTS_KEY_GEN, 0, 0 },
         .sym_keylen = 64,
     },
+    {
+        .name = "Wrap/Unwrap with AES 128 KEY WRAP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 16,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 192 KEY WRAP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 24,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 KEY WRAP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 128 KEY WRAP PAD",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PAD, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 16,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 192 KEY WRAP PAD",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PAD, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 24,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 KEY WRAP PAD",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PAD, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 128 KEY WRAP KWP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_KWP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 16,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 192 KEY WRAP KWP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_KWP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 24,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 KEY WRAP KWP",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_KWP, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 128 KEY WRAP PKCS7",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PKCS7, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 16,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 192 KEY WRAP PKCS7",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PKCS7, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 24,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 KEY WRAP PKCS7",
+        .wrapping_mech = { CKM_AES_KEY_WRAP_PKCS7, NULL, 0 },
+        .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "Wrap/Unwrap with RSA 2048 AES KEY WRAP (AES 256, OAEP SHA-1)",
+        .wrapping_mech = { CKM_RSA_AES_KEY_WRAP,
+                           &rsa_aeskw_aes_256_oaep_sha1_params,
+                           sizeof(CK_RSA_AES_KEY_WRAP_PARAMS) },
+        .wrapping_key_gen_mech = { CKM_RSA_PKCS_KEY_PAIR_GEN, 0, 0 },
+        .rsa_modbits = 2048,
+        .rsa_publ_exp_len = 3,
+        .rsa_publ_exp = {0x01, 0x00, 0x01},
+    },
+    {
+        .name = "Wrap/Unwrap with RSA 4096 AES KEY WRAP (AES 256, OAEP SHA-1)",
+        .wrapping_mech = { CKM_RSA_AES_KEY_WRAP,
+                           &rsa_aeskw_aes_256_oaep_sha1_params,
+                           sizeof(CK_RSA_AES_KEY_WRAP_PARAMS) },
+        .wrapping_key_gen_mech = { CKM_RSA_PKCS_KEY_PAIR_GEN, 0, 0 },
+        .rsa_modbits = 4096,
+        .rsa_publ_exp_len = 3,
+        .rsa_publ_exp = {0x01, 0x00, 0x01},
+    },
+    {
+        .name = "Wrap/Unwrap with ECDH AES KEY WRAP (AES 256, P-256)",
+        .wrapping_mech = { CKM_ECDH_AES_KEY_WRAP,
+                           &ecdh_aeskw_aes_256__params,
+                           sizeof(CK_ECDH_AES_KEY_WRAP_PARAMS) },
+        .wrapping_key_gen_mech = { CKM_EC_KEY_PAIR_GEN, 0, 0 },
+        .ec_parms = prime256v1,
+        .ec_parms_len = sizeof(prime256v1),
+    },
     {
         .name = "Wrap/Unwrap with DES ECB",
         .wrapping_mech = { CKM_DES_ECB, 0, 0 },
@@ -253,8 +398,6 @@ struct wrapping_mech_info wrapping_tests[] = {
 #define NUM_WRAPPING_TESTS sizeof(wrapping_tests) / \
                                 sizeof(struct wrapping_mech_info)
 
-CK_BYTE prime256v1[] = OCK_PRIME256V1;
-
 struct wrapped_mech_info {
     char *name;
     CK_MECHANISM wrapped_key_gen_mech;
@@ -264,6 +407,7 @@ struct wrapped_mech_info {
     CK_ULONG sym_keylen;
     CK_BYTE *ec_params;
     CK_ULONG ec_params_len;
+    CK_ULONG pqc_keyform;
     CK_MECHANISM operation_mech;
 };
 
@@ -350,6 +494,18 @@ struct wrapped_mech_info wrapped_key_tests[] = {
         .ec_params = prime256v1,
         .ec_params_len = sizeof(prime256v1),
     },
+    {
+        .name = "key type IBM Dilithium R2 6-5",
+        .wrapped_key_gen_mech = { CKM_IBM_DILITHIUM, 0, 0 },
+        .operation_mech = { CKM_IBM_DILITHIUM, 0, 0 },
+        .pqc_keyform = CK_IBM_DILITHIUM_KEYFORM_ROUND2_65,
+    },
+    {
+        .name = "key type IBM Dilithium R3 6-5",
+        .wrapped_key_gen_mech = { CKM_IBM_DILITHIUM, 0, 0 },
+        .operation_mech = { CKM_IBM_DILITHIUM, 0, 0 },
+        .pqc_keyform = CK_IBM_DILITHIUM_KEYFORM_ROUND3_65,
+    },
 };
 
 #define NUM_WRAPPED_KEY_TESTS sizeof(wrapped_key_tests) / \
@@ -383,7 +539,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
                                0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,
                                0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f};
     CK_BYTE cipher_data[512];
-    CK_BYTE output_data[512];
+    CK_BYTE output_data[8192];
     CK_OBJECT_HANDLE encr_key, decr_key, sign_key, verify_key;
 
     /* Check if Encrypt/Decrypt or Sign/Verify is supported */
@@ -464,13 +620,13 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
         }
 
         if (output_size != input_size) {
-            testcase_error("Decrypted data has different size then original "
+            testcase_error("Decrypted data has different size than original "
                            "data: orig: %lu decr: %lu", input_size,
                            output_size);
             return CKR_FUNCTION_FAILED;
         }
         if (memcmp(input_data, output_data, input_size) != 0) {
-            testcase_error("Decrypted data is different then original data");
+            testcase_error("Decrypted data is different than original data");
             return CKR_FUNCTION_FAILED;
         }
     } else if (sign && verify) {
@@ -483,6 +639,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
             break;
         case CKM_RSA_PKCS:
         case CKM_ECDSA:
+        case CKM_IBM_DILITHIUM:
             sign_key = priv_key2;
             verify_key = publ_key1;
             input_size = 20;
@@ -503,6 +660,13 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
         rc = funcs->C_Sign(sess2, input_data, input_size, output_data,
                            &output_size);
         if (rc != CKR_OK) {
+            if (mech->mechanism == CKM_IBM_DILITHIUM &&
+                rc == CKR_KEY_SIZE_RANGE) {
+                testcase_skip("IBM Dilithium key form not supported by slot %lu",
+                              slot2);
+                return CKR_OK;
+            }
+
             testcase_error("C_Sign on slot %lu rc=%s", slot2, p11_get_ckr(rc));
             return rc;
         }
@@ -536,8 +700,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     CK_OBJECT_HANDLE priv_key = CK_INVALID_HANDLE;
     CK_OBJECT_HANDLE sym_key = CK_INVALID_HANDLE;
     CK_OBJECT_HANDLE unwrapped_key = CK_INVALID_HANDLE;
-    CK_BYTE wrapped_key[4096];
-    CK_ULONG wrapped_key_size = sizeof(wrapped_key);
+    CK_BYTE *wrapped_key = NULL;
+    CK_ULONG wrapped_key_size = 0;
     CK_OBJECT_CLASS key_class;
     CK_KEY_TYPE key_type;
     CK_ULONG key_len, unwrapped_keylen;
@@ -595,6 +759,22 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         goto testcase_cleanup;
     }
 
+    if (wrap_mech->mechanism == CKM_AES_KEY_WRAP &&
+        tsuite->wrapped_key_gen_mech.mechanism == CKM_DES_KEY_GEN) {
+        testcase_skip("Mechanism CKM_AES_KEY_WRAP can not wrap DES keys "
+                      "(min 16 bytes required)");
+        goto testcase_cleanup;
+    }
+
+    if (wrap_mech->mechanism == CKM_AES_KEY_WRAP &&
+        tsuite->wrapped_key_gen_mech.mechanism == CKM_GENERIC_SECRET_KEY_GEN &&
+        tsuite->sym_keylen < 16) {
+        testcase_skip("Mechanism CKM_AES_KEY_WRAP can not wrap generic secret "
+                      "keys of size %lu (min 16 bytes required)",
+                      tsuite->sym_keylen);
+        goto testcase_cleanup;
+    }
+
     /* Generate the to be wrapped key in slot 1 */
     switch (tsuite->wrapped_key_gen_mech.mechanism) {
     case CKM_RSA_PKCS_KEY_PAIR_GEN:
@@ -664,7 +844,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
 
     case CKM_AES_KEY_GEN:
     case CKM_AES_XTS_KEY_GEN:
-        if ((is_ep11_token(slot_id1) || is_ep11_token(slot_id2)) &&
+        if ((is_ep11_token(slot_id1) || is_ep11_token(slot_id2) ||
+             is_cca_token(slot_id1) || is_cca_token(slot_id2)) &&
             tsuite->wrapped_key_gen_mech.mechanism == CKM_AES_XTS_KEY_GEN) {
             testcase_skip("Skipping as AES XTS is supported only with protected keys");
             rc = CKR_OK;
@@ -692,6 +873,23 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
                                  CK_TRUE); // must be extractable for Wrap/Unwrap
         break;
 
+    case CKM_IBM_DILITHIUM:
+        rc = generate_Dilithium_KeyPair(session1, tsuite->pqc_keyform,
+                                        &publ_key, &priv_key,
+                                        CK_TRUE); // must be extractable for Wrap/Unwrap
+
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("generate to be wrapped key %s with mech %s (%u) in "
+                          "slot %lu is not supported",
+                          tsuite->name,
+                          mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapped_key_gen_mech.mechanism,
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+        break;
+
     default:
         testcase_error("Testcase does not support %s (%u)",
                        mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
@@ -726,17 +924,24 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         goto testcase_cleanup;
 
     /* Wrap the key on slot 1 */
+do_wrap:
     rc = funcs->C_WrapKey(session1, wrap_mech,
                           sym_wrap_key1 != CK_INVALID_HANDLE ?
                                         sym_wrap_key1 : publ_wrap_key1,
                           sym_key != CK_INVALID_HANDLE ? sym_key : priv_key,
                           wrapped_key, &wrapped_key_size);
     if (rc != CKR_OK) {
-        if (rc == CKR_KEY_NOT_WRAPPABLE) {
+        if (rc == CKR_KEY_NOT_WRAPPABLE ||
+            rc == CKR_WRAPPING_KEY_TYPE_INCONSISTENT) {
             testcase_skip("Key not wrappable");
             rc = CKR_OK;
             goto testcase_cleanup;
         }
+        if (rc == CKR_KEY_SIZE_RANGE || rc == CKR_WRAPPING_KEY_SIZE_RANGE) {
+            testcase_skip("Key size not supported");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
         if (rc == CKR_MECHANISM_PARAM_INVALID &&
             wrap_mech->mechanism == CKM_RSA_PKCS_OAEP) {
             oaep = (CK_RSA_PKCS_OAEP_PARAMS *)wrap_mech->pParameter;
@@ -781,6 +986,16 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         goto testcase_cleanup;
     }
 
+    if (wrapped_key == NULL) {
+        wrapped_key = malloc(wrapped_key_size);
+        if (wrapped_key == NULL) {
+            testcase_error("Failed to allocate a buffer of size %lu.",
+                           wrapped_key_size);
+            goto testcase_cleanup;
+        }
+        goto do_wrap;
+    }
+
     /* Get class and key type from original key */
     switch (tsuite->wrapped_key_gen_mech.mechanism) {
     case CKM_AES_KEY_GEN:
@@ -823,11 +1038,17 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
                             wrapped_key, wrapped_key_size, unwrap_tmpl,
                             unwrap_tmpl_num, &unwrapped_key);
     if (rc != CKR_OK) {
-        if (rc == CKR_KEY_NOT_WRAPPABLE || rc == CKR_WRAPPED_KEY_INVALID) {
+        if (rc == CKR_KEY_NOT_WRAPPABLE || rc == CKR_WRAPPED_KEY_INVALID ||
+            rc == CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT) {
             testcase_skip("Key not (un-)wrappable");
             rc = CKR_OK;
             goto testcase_cleanup;
         }
+        if (rc == CKR_KEY_SIZE_RANGE || rc == CKR_UNWRAPPING_KEY_SIZE_RANGE) {
+            testcase_skip("Key size not supported");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
         if (rc == CKR_MECHANISM_PARAM_INVALID &&
             wrap_mech->mechanism == CKM_RSA_PKCS_OAEP) {
             oaep = (CK_RSA_PKCS_OAEP_PARAMS *)wrap_mech->pParameter;
@@ -855,7 +1076,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         }
         if ((rc == CKR_MECHANISM_INVALID || rc == CKR_KEY_TYPE_INCONSISTENT) &&
             is_ep11_token(slot_id2) && wrap_mech->mechanism == CKM_DES3_CBC &&
-            (key_type == CKK_EC || key_type == CKK_RSA)) {
+            (key_type == CKK_EC || key_type == CKK_RSA ||
+             key_type == CKK_IBM_PQC_DILITHIUM)) {
             testcase_skip("EP11 does not support unwrap of EC or RSA keys with DES3 CBC");
             rc = CKR_OK;
             goto testcase_cleanup;
@@ -920,6 +1142,8 @@ testcase_cleanup:
         if (loc_rc != CKR_OK)
             testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(loc_rc));
     }
+    if (wrapped_key != NULL)
+        free(wrapped_key);
 
     if (s != NULL)
         free(s);
@@ -941,10 +1165,16 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     CK_BYTE key[64];
     CK_ULONG key_size = 0;
     CK_BYTE value[64];
+    CK_BYTE ec_params[50];
+    CK_BYTE ec_point[500];
     CK_ATTRIBUTE rsa_publ_tmpl[] = {
         {CKA_MODULUS, modulus, sizeof(modulus) },
         {CKA_PUBLIC_EXPONENT, publ_exp, sizeof(publ_exp) },
     };
+    CK_ATTRIBUTE ec_publ_tmpl[] = {
+        {CKA_EC_PARAMS, ec_params, sizeof(ec_params) },
+        {CKA_EC_POINT, ec_point, sizeof(ec_point) },
+    };
     CK_ATTRIBUTE sym_tmpl[] = {
         {CKA_VALUE, value, sizeof(value) },
     };
@@ -1071,6 +1301,13 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
                                        &publ_wrap_key2, &priv_wrap_key2);
         break;
 
+    case CKM_EC_KEY_PAIR_GEN:
+        rc = generate_EC_KeyPair(session2, tsuite->ec_parms,
+                                 tsuite->ec_parms_len,
+                                 &publ_wrap_key2, &priv_wrap_key2,
+                                 FALSE);
+        break;
+
     case CKM_AES_KEY_GEN:
         key_size = tsuite->sym_keylen;
         rc = funcs->C_GenerateRandom(session2, key, key_size);
@@ -1220,6 +1457,20 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
                            &sym_wrap_key1);
         break;
 
+    case CKM_EC_KEY_PAIR_GEN:
+        rc = funcs->C_GetAttributeValue(session2, publ_wrap_key2,
+                                        ec_publ_tmpl, 2);
+        if (rc != CKR_OK) {
+            testcase_error("C_GetAttributeValue(), rc=%s.", p11_get_ckr(rc));
+            goto testcase_cleanup;
+        }
+        rc = create_ECPublicKey(session1, ec_publ_tmpl[0].pValue,
+                                ec_publ_tmpl[0].ulValueLen,
+                                ec_publ_tmpl[1].pValue,
+                                ec_publ_tmpl[1].ulValueLen, &publ_wrap_key1,
+                                FALSE);
+        break;
+
     default:
         testcase_error("Testcase does not support %s (%u)",
                        mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
@@ -1402,13 +1653,17 @@ int main(int argc, char **argv)
     testcase_pass("C_OpenSession on slot %lu", slot_id2);
 
     testcase_new_assertion();
-    rv = funcs->C_Login(session2, CKU_USER, user_pin, user_pin_len);
-    if (rv != CKR_OK) {
-        testcase_fail("C_Login() on slot %lu rc = %s\n", slot_id2,
-                       p11_get_ckr(rv));
-        // ignore error
-    } else {
+    if (slot_id1 != slot_id2) {
+        rv = funcs->C_Login(session2, CKU_USER, user_pin, user_pin_len);
+        if (rv != CKR_OK) {
+            testcase_fail("C_Login() on slot %lu rc = %s\n", slot_id2,
+                           p11_get_ckr(rv));
+            goto close_session;
+        }
         testcase_pass("C_Login as User on slot %lu", slot_id2);
+    } else {
+        testcase_skip("C_Login as User on slot %lu (already logged in)",
+                      slot_id2);
     }
 
     rv = do_tok2tok_tests();
