@@ -7,19 +7,19 @@
  * found in the file LICENSE file or at
  * https://opensource.org/licenses/cpl1.0.php
  */
-#ifndef _GNU_SOURCE
-#  define _GNU_SOURCE
+
+#if defined(_AIX)
+    const char *__progname = "tableidxgen";
 #endif
+
+#include "platform.h"
 #include <assert.h>
-#include <err.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <time.h>
 #include <unistd.h>
-#include <getopt.h>
 
-#include <pkcs11types.h>
+#include "pkcs11types.h"
 
 #define MECHTABLE_IN_GEN
 #include <mechtable.h>
@@ -64,7 +64,11 @@ const struct aliaslist {
                { "CKM_ECDSA_KEY_PAIR_GEN", "CKM_EC_KEY_PAIR_GEN" },
                { "CKM_IBM_EC_C25519", "CKM_IBM_EC_X25519" },
                { "CKM_IBM_EC_C448", "CKM_IBM_EC_X448" },
-               { "CKM_IBM_EDDSA_SHA512", "CKM_IBM_ED25519_SHA512" }
+               { "CKM_IBM_EDDSA_SHA512", "CKM_IBM_ED25519_SHA512" },
+               { "CKM_SHA3_224_KEY_DERIVE", "CKM_SHA3_224_KEY_DERIVATION" },
+               { "CKM_SHA3_256_KEY_DERIVE", "CKM_SHA3_256_KEY_DERIVATION" },
+               { "CKM_SHA3_384_KEY_DERIVE", "CKM_SHA3_384_KEY_DERIVATION" },
+               { "CKM_SHA3_512_KEY_DERIVE", "CKM_SHA3_512_KEY_DERIVATION" },
 };
 
 /* Get the table */
@@ -581,18 +585,8 @@ static void dumpstringfun(FILE *fp)
 
 static void generatelicense(FILE *fp)
 {
-    time_t t;
-    struct tm *tm;
-    char *tstr;
-
-    t = time(NULL);
-    tm = localtime(&t);
-    tstr = asctime(tm);
-    /* Remove trailing newline character */
-    tstr[strlen(tstr) - 1] = 0;
     fputs("/*\n", fp);
-    fprintf(fp, " * COPYRIGHT (c) International Business Machines Corp. %d\n",
-            tm->tm_year + 1900);
+    fprintf(fp, " * COPYRIGHT (c) International Business Machines Corp. 2024\n");
     fputs(" *\n", fp);
     fputs(" * This program is provided under the terms of the Common Public License,\n",
           fp);
@@ -603,7 +597,7 @@ static void generatelicense(FILE *fp)
     fputs(" * found in the file LICENSE file or at\n", fp);
     fputs(" * https://opensource.org/licenses/cpl1.0.php\n", fp);
     fputs(" */\n", fp);
-    fprintf(fp, "/* AUTO-GENERATED on %s.  DO NOT EDIT! */\n\n", tstr);
+    fprintf(fp, "/* AUTO-GENERATED.  DO NOT EDIT! */\n\n");
 }
 
 static void generateheader(char *hname)
