@@ -33,10 +33,24 @@
 #define OCK_HSM_MK_CHANGE_PATH CONFIG_PATH "/HSM_MK_CHANGE"
 #define OCK_HSM_MK_CHANGE_LOCK_FILE LOCKDIR_PATH "/LCK..HSM_MK_CHANGElock"
 
-#define PROC_SOCKET_FILE_PATH "/run/opencryptoki/pkcsslotd.socket"
-#define ADMIN_SOCKET_FILE_PATH "/run/opencryptoki/pkcsslotd.admin.socket"
+#if defined(_AIX)
+   /*
+    * AIX needs the file to be opened in write mode for an exclusive lock.
+    * Mode bits are changed to -w--w---- to allow starting up even when there
+    * is a lock file left over from a previous start.
+    */
+    #define MODE_BITS (S_IWUSR | S_IWGRP)
+    #define OPEN_MODE O_WRONLY
+#else
+    #define MODE_BITS (S_IRUSR | S_IRGRP)
+    #define OPEN_MODE O_RDONLY
+#endif
 
-#define PID_FILE_PATH "/run/opencryptoki/pkcsslotd.pid"
+#define RUN_DIR RUN_PATH "/opencryptoki"
+#define PROC_SOCKET_FILE_PATH RUN_DIR "/pkcsslotd.socket"
+#define ADMIN_SOCKET_FILE_PATH RUN_DIR "/pkcsslotd.admin.socket"
+
+#define PID_FILE_PATH RUN_DIR "/pkcsslotd.pid"
 #define OCK_CONFIG OCK_CONFDIR "/opencryptoki.conf"
 
 #ifndef PKCSSLOTD_USER
@@ -59,6 +73,16 @@
 #define NUMBER_PROCESSES_ALLOWED  1000
 #define NUMBER_ADMINS_ALLOWED     1000
 
+#if defined(_AIX)
+   /*
+    * AIX does not define NAME_MAX and instead requires this
+    * to be determined at runtime. Use the equivalent defined for
+    * BSD-compatibility (which uses MAXNAMLEN) - also 255
+    */
+    #include <dirent.h>
+    #define NAME_MAX _D_NAME_MAX
+#endif
+
 //
 // Per Process Data structure
 // one entry in the table is grabbed by each process
@@ -205,6 +229,7 @@ typedef struct {
     char tokname[NAME_MAX + 1]; // token specific directory
     LW_SHM_TYPE *shm_addr;      // token specific shm address
     uint32_t version; // version: major<<16|minor
+    char usergroup[LOGIN_NAME_MAX]; // group of users having access to the token
 } Slot_Info_t_64;
 
 typedef Slot_Info_t_64 SLOT_INFO;
