@@ -29,6 +29,7 @@ typedef struct {
     struct bt_ref_hdr hdr;
     CK_SLOT_ID slotID;
     CK_SESSION_HANDLE sessionh;
+    CK_BBOOL rw_session;
 } ST_SESSION_T;
 
 typedef struct trace_handle_t trace_handle;
@@ -338,6 +339,9 @@ typedef CK_RV (CK_PTR ST_C_WaitForSlotEvent) (STDLL_TokData_t *tokdata,
                                               CK_FLAGS flags,
                                               CK_SLOT_ID_PTR pSlot,
                                               CK_VOID_PTR pReserved);
+typedef CK_RV (CK_PTR ST_C_SessionCancel)(STDLL_TokData_t *tokdata,
+                                          ST_SESSION_T *hSession,
+                                          CK_FLAGS flags);
 
 typedef CK_RV (CK_PTR ST_C_IBM_ReencryptSingle)(STDLL_TokData_t *tokdata,
                                                 ST_SESSION_T *hSession,
@@ -350,6 +354,11 @@ typedef CK_RV (CK_PTR ST_C_IBM_ReencryptSingle)(STDLL_TokData_t *tokdata,
                                                 CK_BYTE_PTR pReencryptedData,
                                             CK_ULONG_PTR pulReencryptedDataLen);
 
+typedef CK_RV (CK_PTR ST_C_HandleEvent)(STDLL_TokData_t *tokdata,
+                                        unsigned int event_type,
+                                        unsigned int event_flags,
+                                        const char *payload,
+                                        unsigned int payload_len);
 
 struct ST_FCN_LIST {
 
@@ -422,8 +431,12 @@ struct ST_FCN_LIST {
     // Question if these have to be implemented for Netscape support
     ST_C_GetFunctionStatus ST_GetFunctionStatus;
     ST_C_CancelFunction ST_CancelFunction;
+    ST_C_SessionCancel ST_SessionCancel;
 
     ST_C_IBM_ReencryptSingle ST_IBM_ReencryptSingle;
+
+    /* The functions defined below are not part of the external API */
+    ST_C_HandleEvent ST_HandleEvent;
 };
 
 typedef struct ST_FCN_LIST STDLL_FcnList_t;
