@@ -578,7 +578,7 @@ typedef void (*CSNDPKX_t) (long *return_code,
                            long *target_key_token_length,
                            unsigned char *target_key_token);
 
-/* PKA Symmetric Key Import */
+/* Symmetric Key Import */
 typedef void (*CSNDSYI_t) (long *return_code,
                            long *reason_code,
                            long *exit_data_length,
@@ -592,7 +592,23 @@ typedef void (*CSNDSYI_t) (long *return_code,
                            long *target_key_identifier_length,
                            unsigned char *target_key_identifier);
 
-/* PKA Symmetric Key Export */
+/* Symmetric Key Import 2 */
+typedef void (*CSNDSYI2_t) (long *return_code,
+                            long *reason_code,
+                            long *exit_data_length,
+                            unsigned char *exit_data,
+                            long *rule_array_count,
+                            unsigned char *rule_array,
+                            long *enciphered_key_length,
+                            unsigned char *enciphered_key,
+                            long *transport_key_length,
+                            unsigned char *transport_key,
+                            long *key_name_len,
+                            unsigned char *key_name,
+                            long *target_key_identifier_length,
+                            unsigned char *target_key_identifier);
+
+/* Symmetric Key Export */
 typedef void (*CSNDSYX_t) (long *return_code,
                            long *reason_code,
                            long *exit_data_length,
@@ -1139,20 +1155,6 @@ typedef void (*CSNBPCU_t) (long *return_code,
                            long *outputPinMessageLength,
                            unsigned char *outputPinMessage);
 
-/* PCF/CUSP Key Conversion */
-typedef void (*CSUAPCV_t) (long *return_code,
-                           long *reason_code,
-                           long *exit_data_length,
-                           unsigned char *exit_data,
-                           long *rule_array_count,
-                           unsigned char *rule_array,
-                           long *KEK_key_identifier_length,
-                           unsigned char *KEK_key_identifier,
-                           long *PCF_key_list_length,
-                           unsigned char *PCF_key_list,
-                           long *output_key_list_length,
-                           unsigned char *output_key_list);
-
 /*Process Request Block*/
 typedef void (*CSUAPRB_t) (long *pReturnCode,
                            long *pReasonCode,
@@ -1166,38 +1168,6 @@ typedef void (*CSUAPRB_t) (long *pReturnCode,
                            unsigned char *pOutFileName,
                            long *pReplyLength, unsigned char *pReply);
 
-/* Diffie-Hellman Key Load */
-typedef void (*CSUADHK_t) (long *ReturnCode,
-                           long *ReasonCode,
-                           long *ExitDataLength,
-                           unsigned char *ExitData,
-                           long *RuleArrayCount,
-                           unsigned char *RuleArray,
-                           unsigned char *DHModulus,
-                           unsigned char *DHGenerator,
-                           unsigned char *DHKeyPart,
-                           long *TransportKeyHashLength,
-                           unsigned char *TransportKeyHash,
-                           unsigned char *Reserved1,
-                           unsigned char *Reserved2,
-                           unsigned char *Reserved3, unsigned char *Reserved4);
-
-/* Diffie-Hellman Key Query */
-typedef void (*CSUADHQ_t) (long *ReturnCode,
-                           long *ReasonCode,
-                           long *ExitDataLength,
-                           unsigned char *ExitData,
-                           long *RuleArrayCount,
-                           unsigned char *RuleArray,
-                           unsigned char *DHModulus,
-                           unsigned char *DHGenerator,
-                           unsigned char *DHKeyPart,
-                           long *TransportKeyHashLength,
-                           unsigned char *TransportKeyHash,
-                           unsigned char *Reserved1,
-                           unsigned char *Reserved2,
-                           unsigned char *Reserved3, unsigned char *Reserved4);
-
 /* Trusted Block Create */
 typedef void (*CSNDTBC_t) (long *return_code,
                            long *reason_code,
@@ -1320,3 +1290,68 @@ typedef void (*CSUACFV_t)(long *return_code,
                           unsigned char *exit_data,
                           long *version_data_length,
                           unsigned char *version_data);
+
+/* Restrict Key Attribute */
+typedef void (*CSNBRKA_t)(long *return_code,
+                          long *reason_code,
+                          long *exit_data_length,
+                          unsigned char *exit_data,
+                          long *rule_array_count,
+                          unsigned char *rule_array,
+                          long *key_identifier_length,
+                          unsigned char *key_identifier,
+                          long *ey_encrypting_key_identifier_length,
+                          unsigned char *ey_encrypting_key_identifier,
+                          long *opt_parameter1_length,
+                          unsigned char *opt_parameter1,
+                          long *opt_parameter2_length,
+                          unsigned char *opt_parameter2);
+
+/* Key Translate2 */
+typedef void (*CSNBKTR2_t)(long *return_code,
+                           long *reason_code,
+                           long *exit_data_length,
+                           unsigned char *exit_data,
+                           long *rule_array_count,
+                           unsigned char *rule_array,
+                           long *input_key_token_length,
+                           unsigned char *input_key_token,
+                           long *input_KEK_key_identifier_length,
+                           unsigned char *input_KEK_key_identifier,
+                           long *output_KEK_key_identifier_length,
+                           unsigned char *output_KEK_key_identifier,
+                           long *output_key_token_length,
+                           unsigned char *output_key_token);
+
+/* CCA EC Diffie-Hellman function */
+typedef void (*CSNDEDH_t)(long *return_code,
+                          long *reason_code,
+                          long *exit_data_length,
+                          unsigned char *exit_data,
+                          long *rule_array_count,
+                          unsigned char *rule_array,
+                          long *private_key_identifier_length,
+                          unsigned char *private_key_identifier,
+                          long *private_KEK_key_identifier_length,
+                          unsigned char *private_KEK_key_identifier,
+                          long *public_key_identifier_length,
+                          unsigned char *public_key_identifier,
+                          long *chaining_vector_length,
+                          unsigned char *chaining_vector,
+                          long *party_info_length,
+                          unsigned char *party_info,
+                          long *key_bit_length,
+                          long *reserved_1_length,
+                          unsigned char *reserved_1,
+                          long *reserved_2_length,
+                          unsigned char *reserved_2,
+                          long *reserved_3_length,
+                          unsigned char *reserved_3,
+                          long *reserved_4_length,
+                          unsigned char *reserved_4,
+                          long *reserved_5_length,
+                          unsigned char *reserved_5,
+                          long *output_KEK_key_identifier_length,
+                          unsigned char *output_KEK_key_identifier,
+                          long *output_key_identifier_length,
+                          unsigned char *output_key_identifier);
