@@ -86,6 +86,7 @@ token_spec_t token_specific = {
     &token_specific_ec_verify,
     &token_specific_ec_generate_keypair,
     NULL,                       // ecdh_derive
+    &token_specific_ecdh_pkcs_derive_kdf,
     NULL,                       // dh_pkcs_derive
     NULL,                       // dh_pkcs_key_pair_gen
     // SHA
@@ -93,6 +94,8 @@ token_spec_t token_specific = {
     token_specific_sha,
     token_specific_sha_update,
     token_specific_sha_final,
+    // SHAKE derive
+    NULL,                       // shake_key_derive
     // HMAC
     &token_specific_hmac_sign_init,
     &token_specific_hmac_sign,
@@ -114,8 +117,8 @@ token_spec_t token_specific = {
     &token_specific_aes_ecb,
     &token_specific_aes_cbc,
     NULL,                       // aes_ctr
-    NULL,                       // aes_gcm_init,
-    NULL,                       // aes_gcm
+    &token_specific_aes_gcm_init,
+    &token_specific_aes_gcm,
     NULL,                       // aes_gcm_update
     NULL,                       // aes_gcm_final
     NULL,                       // aes_ofb
@@ -127,22 +130,23 @@ token_spec_t token_specific = {
 #else
     NULL,
 #endif
+    NULL,                       // aes_key_wrap
     // DSA
     NULL,                       // dsa_generate_keypair
     NULL,                       // dsa_sign
     NULL,                       // dsa_verify
+    // PQC
+    &token_specific_ibm_dilithium_generate_keypair,
+    &token_specific_ibm_dilithium_sign,
+    &token_specific_ibm_dilithium_verify,
     &token_specific_get_mechanism_list,
     &token_specific_get_mechanism_info,
     &token_specific_object_add,
     &token_specific_key_wrap,
     &token_specific_key_unwrap,
     &token_specific_reencrypt_single,
-    NULL,                       // set_attribute_values
-#ifndef NO_PKEY
+    &token_specific_set_attribute_values,
     &token_specific_set_attrs_for_new_object,
-#else
-    NULL,
-#endif
     &token_specific_handle_event,
     NULL,                       // check_obj_access
 };
