@@ -34,7 +34,7 @@ CK_RV decr_mgr_init(STDLL_TokData_t *tokdata,
                     ENCR_DECR_CONTEXT *ctx,
                     CK_ULONG operation,
                     CK_MECHANISM *mech, CK_OBJECT_HANDLE key_handle,
-                    CK_BBOOL checkpolicy)
+                    CK_BBOOL checkpolicy, CK_BBOOL checkauth)
 {
     OBJECT *key_obj = NULL;
     CK_BYTE *ptr = NULL;
@@ -45,6 +45,7 @@ CK_RV decr_mgr_init(STDLL_TokData_t *tokdata,
     CK_ULONG strength = POLICY_STRENGTH_IDX_0;
     CK_GCM_PARAMS gcm_params;
     CK_MECHANISM temp_mech;
+    CK_ULONG aeskw_iv_len = AES_KEY_WRAP_KWP_IV_SIZE;
 
     if (!sess) {
         TRACE_ERROR("Invalid function arguments.\n");
@@ -118,6 +119,16 @@ CK_RV decr_mgr_init(STDLL_TokData_t *tokdata,
         }
     }
 
+    ctx->auth_required = FALSE;
+    if (checkauth) {
+        rc = key_object_is_always_authenticate(key_obj->template,
+                                               &ctx->auth_required);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("key_object_is_always_authenticate failed\n");
+            goto done;
+        }
+    }
+
     if (!key_object_is_mechanism_allowed(key_obj->template, mech->mechanism)) {
         TRACE_ERROR("Mechanism not allowed per CKA_ALLOWED_MECHANISMS.\n");
         rc = CKR_MECHANISM_INVALID;
@@ -361,6 +372,70 @@ CK_RV decr_mgr_init(STDLL_TokData_t *tokdata,
             goto done;
         }
 
+        // RSA cannot be used for multi-part operations
+        //
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+    case CKM_ECDH_AES_KEY_WRAP:
+        /* CKM_ECDH_AES_KEY_WRAP can only be used for wrap/unwrap */
+        if (operation != OP_UNWRAP) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
+            rc = CKR_MECHANISM_INVALID;
+            goto done;
+        }
+
+        if (mech->ulParameterLen != sizeof(CK_ECDH_AES_KEY_WRAP_PARAMS) ||
+            mech->pParameter == NULL) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_EC) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+        // cannot be used for multi-part operations
+        //
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+    case CKM_RSA_AES_KEY_WRAP:
+        /* CKM_RSA_AES_KEY_WRAP can only be used for wrap/unwrap */
+        if (operation != OP_UNWRAP) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
+            rc = CKR_MECHANISM_INVALID;
+            goto done;
+        }
+
+        if (mech->ulParameterLen != sizeof(CK_RSA_AES_KEY_WRAP_PARAMS) ||
+            mech->pParameter == NULL) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_RSA) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
         // RSA cannot be used for multi-part operations
         //
         ctx->context_len = 0;
@@ -576,6 +651,38 @@ CK_RV decr_mgr_init(STDLL_TokData_t *tokdata,
         }
         memset(ctx->context, 0x0, sizeof(AES_XTS_CONTEXT));
         break;
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_PKCS7:
+        aeskw_iv_len = AES_KEY_WRAP_IV_SIZE;
+        /* fallthrough */
+    case CKM_AES_KEY_WRAP_KWP:
+        if ((mech->ulParameterLen != 0 && mech->pParameter == NULL) ||
+            (mech->ulParameterLen == 0 && mech->pParameter != NULL) ||
+            (mech->ulParameterLen != 0 &&
+             mech->ulParameterLen != aeskw_iv_len)) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_AES) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+
+        /* AES key wrap cannot be used for multi-part operations */
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
         rc = CKR_MECHANISM_INVALID;
@@ -643,10 +750,10 @@ CK_RV decr_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
     ctx->multi = FALSE;
     ctx->active = FALSE;
     ctx->init_pending = FALSE;
-    ctx->context_len = 0;
     ctx->pkey_active = FALSE;
     ctx->state_unsaveable = FALSE;
     ctx->count_statistics = FALSE;
+    ctx->auth_required = FALSE;
 
     if (ctx->mech.pParameter) {
         /* Deep free mechanism parameter, if required */
@@ -673,6 +780,7 @@ CK_RV decr_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
             free(ctx->context);
         ctx->context = NULL;
     }
+    ctx->context_len = 0;
     ctx->context_free_func = NULL;
 
     return CKR_OK;
@@ -698,6 +806,10 @@ CK_RV decr_mgr_decrypt(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_OPERATION_NOT_INITIALIZED));
         return CKR_OPERATION_NOT_INITIALIZED;
     }
+    if (ctx->auth_required == TRUE) {
+        TRACE_ERROR("%s\n", ock_err(ERR_USER_NOT_LOGGED_IN));
+        return CKR_USER_NOT_LOGGED_IN;
+    }
     if (ctx->multi_init == FALSE) {
         ctx->multi = FALSE;
         ctx->multi_init = TRUE;
@@ -785,6 +897,14 @@ CK_RV decr_mgr_decrypt(STDLL_TokData_t *tokdata,
         return rsa_x509_decrypt(tokdata, sess, length_only,
                                 ctx,
                                 in_data, in_data_len, out_data, out_data_len);
+    case CKM_ECDH_AES_KEY_WRAP:
+        return ecdh_aes_key_unwrap(tokdata, sess, length_only, ctx,
+                                   in_data, in_data_len,
+                                   out_data, out_data_len);
+    case CKM_RSA_AES_KEY_WRAP:
+        return rsa_aes_key_unwrap(tokdata, sess, length_only, ctx,
+                                  in_data, in_data_len,
+                                  out_data, out_data_len);
     case CKM_AES_CBC:
         return aes_cbc_decrypt(tokdata, sess, length_only,
                                ctx,
@@ -825,9 +945,16 @@ CK_RV decr_mgr_decrypt(STDLL_TokData_t *tokdata,
                                in_data, in_data_len,
                                out_data, out_data_len, 0x10);
     case CKM_AES_XTS:
-            return aes_xts_decrypt(tokdata, sess, length_only, ctx,
-                                   in_data, in_data_len,
-                                   out_data, out_data_len);
+        return aes_xts_decrypt(tokdata, sess, length_only, ctx,
+                               in_data, in_data_len,
+                               out_data, out_data_len);
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_KWP:
+    case CKM_AES_KEY_WRAP_PKCS7:
+        return aes_key_wrap_decrypt(tokdata, sess, length_only, ctx,
+                                    in_data, in_data_len,
+                                    out_data, out_data_len);
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -864,6 +991,10 @@ CK_RV decr_mgr_decrypt_update(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_OPERATION_NOT_INITIALIZED));
         return CKR_OPERATION_NOT_INITIALIZED;
     }
+    if (ctx->auth_required == TRUE) {
+        TRACE_ERROR("%s\n", ock_err(ERR_USER_NOT_LOGGED_IN));
+        return CKR_USER_NOT_LOGGED_IN;
+    }
     if (ctx->multi_init == FALSE) {
         ctx->multi = TRUE;
         ctx->multi_init = TRUE;
@@ -1013,6 +1144,10 @@ CK_RV decr_mgr_decrypt_final(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_OPERATION_NOT_INITIALIZED));
         return CKR_OPERATION_NOT_INITIALIZED;
     }
+    if (ctx->auth_required == TRUE) {
+        TRACE_ERROR("%s\n", ock_err(ERR_USER_NOT_LOGGED_IN));
+        return CKR_USER_NOT_LOGGED_IN;
+    }
     if (ctx->multi_init == FALSE) {
         ctx->multi = TRUE;
         ctx->multi_init = TRUE;
