@@ -45,6 +45,9 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
     CK_RV rc;
     int check;
     CK_ULONG strength = POLICY_STRENGTH_IDX_0;
+    CK_GCM_PARAMS gcm_params;
+    CK_MECHANISM temp_mech;
+    CK_ULONG aeskw_iv_len = AES_KEY_WRAP_KWP_IV_SIZE;
 
     if (!sess || !ctx || !mech) {
         TRACE_ERROR("Invalid function arguments.\n");
@@ -159,36 +162,6 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
         //       rc = CKR_MECHANISM_INVALID;
         //       goto done;
 
-        ctx->context_len = sizeof(DES_CONTEXT);
-        ctx->context = (CK_BYTE *) malloc(sizeof(DES_CONTEXT));
-        if (!ctx->context) {
-            TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-            rc = CKR_HOST_MEMORY;
-            goto done;
-        }
-        memset(ctx->context, 0x0, sizeof(DES_CONTEXT));
-        break;
-    case CKM_CDMF_ECB:
-        if (mech->ulParameterLen != 0) {
-            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-            rc = CKR_MECHANISM_PARAM_INVALID;
-            goto done;
-        }
-        // is the key type correct?
-        //
-        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
-                                          &keytype);
-        if (rc != CKR_OK) {
-            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
-            goto done;
-        }
-
-        if (keytype != CKK_CDMF) {
-            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
-            rc = CKR_KEY_TYPE_INCONSISTENT;
-            goto done;
-        }
-
         ctx->context_len = sizeof(DES_CONTEXT);
         ctx->context = (CK_BYTE *) malloc(sizeof(DES_CONTEXT));
         if (!ctx->context) {
@@ -260,38 +233,6 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
             goto done;
         }
 
-        ctx->context_len = sizeof(DES_CONTEXT);
-        ctx->context = (CK_BYTE *) malloc(sizeof(DES_CONTEXT));
-        if (!ctx->context) {
-            TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-            rc = CKR_HOST_MEMORY;
-            goto done;
-        }
-        memset(ctx->context, 0x0, sizeof(DES_CONTEXT));
-        break;
-    case CKM_CDMF_CBC:
-    case CKM_CDMF_CBC_PAD:
-        if (mech->ulParameterLen != DES_BLOCK_SIZE ||
-            mech->pParameter == NULL) {
-            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-            rc = CKR_MECHANISM_PARAM_INVALID;
-            goto done;
-        }
-        // is the key type correct?
-        //
-        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
-                                          &keytype);
-        if (rc != CKR_OK) {
-            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
-            goto done;
-        }
-
-        if (keytype != CKK_CDMF) {
-            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
-            rc = CKR_KEY_TYPE_INCONSISTENT;
-            goto done;
-        }
-
         ctx->context_len = sizeof(DES_CONTEXT);
         ctx->context = (CK_BYTE *) malloc(sizeof(DES_CONTEXT));
         if (!ctx->context) {
@@ -423,6 +364,70 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
             goto done;
         }
 
+        // RSA cannot be used for multi-part operations
+        //
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+    case CKM_ECDH_AES_KEY_WRAP:
+        /* CKM_ECDH_AES_KEY_WRAP can only be used for wrap/unwrap */
+        if (operation != OP_WRAP) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
+            rc = CKR_MECHANISM_INVALID;
+            goto done;
+        }
+
+        if (mech->ulParameterLen != sizeof(CK_ECDH_AES_KEY_WRAP_PARAMS) ||
+            mech->pParameter == NULL) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_EC) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+        // cannot be used for multi-part operations
+        //
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+    case CKM_RSA_AES_KEY_WRAP:
+        /* CKM_RSA_AES_KEY_WRAP can only be used for wrap/unwrap */
+        if (operation != OP_WRAP) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
+            rc = CKR_MECHANISM_INVALID;
+            goto done;
+        }
+
+        if (mech->ulParameterLen != sizeof(CK_RSA_AES_KEY_WRAP_PARAMS) ||
+            mech->pParameter == NULL) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_RSA) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
         // RSA cannot be used for multi-part operations
         //
         ctx->context_len = 0;
@@ -523,13 +528,23 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
         memset(ctx->context, 0x0, sizeof(AES_CONTEXT));
         break;
     case CKM_AES_GCM:
-        if (mech->ulParameterLen != sizeof(CK_GCM_PARAMS) ||
+        if ((mech->ulParameterLen != sizeof(CK_GCM_PARAMS) &&
+             mech->ulParameterLen != sizeof(CK_GCM_PARAMS_COMPAT)) ||
             mech->pParameter == NULL) {
             TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
             rc = CKR_MECHANISM_PARAM_INVALID;
             goto done;
         }
 
+        if (mech->ulParameterLen == sizeof(CK_GCM_PARAMS_COMPAT)) {
+            aes_gcm_param_from_compat((CK_GCM_PARAMS_COMPAT *)mech->pParameter,
+                                      &gcm_params);
+            temp_mech.mechanism = mech->mechanism;
+            temp_mech.pParameter = &gcm_params;
+            temp_mech.ulParameterLen = sizeof(gcm_params);
+            mech = &temp_mech;
+        }
+
         rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
                                           &keytype);
         if (rc != CKR_OK) {
@@ -598,6 +613,68 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
         }
         memset(ctx->context, 0x0, sizeof(AES_CONTEXT));
         break;
+    case CKM_AES_XTS:
+        if (mech->ulParameterLen != AES_INIT_VECTOR_SIZE ||
+            mech->pParameter == NULL) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_AES_XTS) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+
+        ctx->context_len = sizeof(AES_XTS_CONTEXT);
+        ctx->context = (CK_BYTE *) malloc(sizeof(AES_XTS_CONTEXT));
+        if (!ctx->context) {
+            TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            rc = CKR_HOST_MEMORY;
+            goto done;
+        }
+        memset(ctx->context, 0x0, sizeof(AES_XTS_CONTEXT));
+        break;
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_PKCS7:
+        aeskw_iv_len = AES_KEY_WRAP_IV_SIZE;
+        /* fallthrough */
+    case CKM_AES_KEY_WRAP_KWP:
+        if ((mech->ulParameterLen != 0 && mech->pParameter == NULL) ||
+            (mech->ulParameterLen == 0 && mech->pParameter != NULL) ||
+            (mech->ulParameterLen != 0 &&
+             mech->ulParameterLen != aeskw_iv_len)) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_AES) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+
+        /* AES key wrap cannot be used for multi-part operations */
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
         rc = CKR_MECHANISM_INVALID;
@@ -621,6 +698,7 @@ CK_RV encr_mgr_init(STDLL_TokData_t *tokdata,
                                    (CK_GCM_PARAMS *)ptr);
             if (rc != CKR_OK) {
                 TRACE_ERROR("aes_gcm_dup_param failed\n");
+                free(ptr);
                 goto done;
             }
             break;
@@ -664,7 +742,6 @@ CK_RV encr_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
     ctx->multi = FALSE;
     ctx->active = FALSE;
     ctx->init_pending = FALSE;
-    ctx->context_len = 0;
     ctx->pkey_active = FALSE;
     ctx->state_unsaveable = FALSE;
     ctx->count_statistics = FALSE;
@@ -694,6 +771,7 @@ CK_RV encr_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
             free(ctx->context);
         ctx->context = NULL;
     }
+    ctx->context_len = 0;
     ctx->context_free_func = NULL;
 
     return CKR_OK;
@@ -736,18 +814,15 @@ CK_RV encr_mgr_encrypt(STDLL_TokData_t *tokdata,
         return CKR_OPERATION_ACTIVE;
     }
     switch (ctx->mech.mechanism) {
-    case CKM_CDMF_ECB:
     case CKM_DES_ECB:
         return pk_des_ecb_encrypt(tokdata, sess, length_only, ctx,
                                   in_data, in_data_len,
                                   out_data, out_data_len);
-    case CKM_CDMF_CBC:
     case CKM_DES_CBC:
         return pk_des_cbc_encrypt(tokdata, sess, length_only, ctx,
                                   in_data, in_data_len,
                                   out_data, out_data_len);
     case CKM_DES_CBC_PAD:
-    case CKM_CDMF_CBC_PAD:
         return des_cbc_pad_encrypt(tokdata, sess, length_only, ctx,
                                    in_data, in_data_len,
                                    out_data, out_data_len);
@@ -805,7 +880,14 @@ CK_RV encr_mgr_encrypt(STDLL_TokData_t *tokdata,
         return rsa_x509_encrypt(tokdata, sess, length_only, ctx,
                                 in_data, in_data_len,
                                 out_data, out_data_len);
-#ifndef NOAES
+    case CKM_ECDH_AES_KEY_WRAP:
+        return ecdh_aes_key_wrap(tokdata, sess, length_only, ctx,
+                                 in_data, in_data_len,
+                                 out_data, out_data_len);
+    case CKM_RSA_AES_KEY_WRAP:
+        return rsa_aes_key_wrap(tokdata, sess, length_only, ctx,
+                                in_data, in_data_len,
+                                out_data, out_data_len);
     case CKM_AES_CBC:
         return aes_cbc_encrypt(tokdata, sess, length_only, ctx,
                                in_data, in_data_len,
@@ -844,7 +926,17 @@ CK_RV encr_mgr_encrypt(STDLL_TokData_t *tokdata,
         return aes_cfb_encrypt(tokdata, sess, length_only, ctx,
                                in_data, in_data_len,
                                out_data, out_data_len, 0x10);
-#endif
+    case CKM_AES_XTS:
+        return aes_xts_encrypt(tokdata, sess, length_only, ctx,
+                               in_data, in_data_len,
+                               out_data, out_data_len);
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_KWP:
+    case CKM_AES_KEY_WRAP_PKCS7:
+        return aes_key_wrap_encrypt(tokdata, sess, length_only, ctx,
+                                    in_data, in_data_len,
+                                    out_data, out_data_len);
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -891,18 +983,15 @@ CK_RV encr_mgr_encrypt_update(STDLL_TokData_t *tokdata,
     }
 
     switch (ctx->mech.mechanism) {
-    case CKM_CDMF_ECB:
     case CKM_DES_ECB:
         return des_ecb_encrypt_update(tokdata, sess, length_only, ctx,
                                       in_data, in_data_len,
                                       out_data, out_data_len);
-    case CKM_CDMF_CBC:
     case CKM_DES_CBC:
         return des_cbc_encrypt_update(tokdata, sess, length_only, ctx,
                                       in_data, in_data_len,
                                       out_data, out_data_len);
     case CKM_DES_CBC_PAD:
-    case CKM_CDMF_CBC_PAD:
         return des_cbc_pad_encrypt_update(tokdata, sess, length_only, ctx,
                                           in_data, in_data_len,
                                           out_data, out_data_len);
@@ -948,7 +1037,6 @@ CK_RV encr_mgr_encrypt_update(STDLL_TokData_t *tokdata,
         return des3_cbc_pad_encrypt_update(tokdata, sess, length_only, ctx,
                                            in_data, in_data_len,
                                            out_data, out_data_len);
-#ifndef NOAES
     case CKM_AES_ECB:
         return aes_ecb_encrypt_update(tokdata, sess, length_only, ctx,
                                       in_data, in_data_len,
@@ -985,7 +1073,10 @@ CK_RV encr_mgr_encrypt_update(STDLL_TokData_t *tokdata,
         return aes_cfb_encrypt_update(tokdata, sess, length_only, ctx,
                                       in_data, in_data_len,
                                       out_data, out_data_len, 0x10);
-#endif
+    case CKM_AES_XTS:
+        return aes_xts_encrypt_update(tokdata, sess, length_only, ctx,
+                                      in_data, in_data_len,
+                                      out_data, out_data_len);
     default:
         return CKR_MECHANISM_INVALID;
     }
@@ -1023,16 +1114,13 @@ encr_mgr_encrypt_final(STDLL_TokData_t *tokdata,
         return CKR_OPERATION_ACTIVE;
     }
     switch (ctx->mech.mechanism) {
-    case CKM_CDMF_ECB:
     case CKM_DES_ECB:
         return des_ecb_encrypt_final(tokdata, sess, length_only,
                                      ctx, out_data, out_data_len);
-    case CKM_CDMF_CBC:
     case CKM_DES_CBC:
         return des_cbc_encrypt_final(tokdata, sess, length_only,
                                      ctx, out_data, out_data_len);
     case CKM_DES_CBC_PAD:
-    case CKM_CDMF_CBC_PAD:
         return des_cbc_pad_encrypt_final(tokdata, sess, length_only,
                                          ctx, out_data, out_data_len);
     case CKM_DES_OFB64:
@@ -1071,7 +1159,6 @@ encr_mgr_encrypt_final(STDLL_TokData_t *tokdata,
     case CKM_DES3_CBC_PAD:
         return des3_cbc_pad_encrypt_final(tokdata, sess, length_only,
                                           ctx, out_data, out_data_len);
-#ifndef NOAES
     case CKM_AES_ECB:
         return aes_ecb_encrypt_final(tokdata, sess, length_only,
                                      ctx, out_data, out_data_len);
@@ -1099,7 +1186,9 @@ encr_mgr_encrypt_final(STDLL_TokData_t *tokdata,
     case CKM_AES_CFB128:
         return aes_cfb_encrypt_final(tokdata, sess, length_only,
                                      ctx, out_data, out_data_len, 0x10);
-#endif
+    case CKM_AES_XTS:
+        return aes_xts_encrypt_final(tokdata, sess, length_only,
+                                     ctx, out_data, out_data_len);
     default:
         return CKR_MECHANISM_INVALID;
     }
@@ -1236,7 +1325,7 @@ CK_RV encr_mgr_reencrypt_single(STDLL_TokData_t *tokdata, SESSION *sess,
     /* No token specific reencrypt_single function, perform it manually */
     /* Enforces policy */
     rc = decr_mgr_init(tokdata, sess, decr_ctx, OP_DECRYPT_INIT, decr_mech,
-                       decr_key, TRUE);
+                       decr_key, TRUE, TRUE);
     if (rc != CKR_OK)
         goto done;
     /* Enforces Policy */
