@@ -95,6 +95,38 @@ CK_RV key_mgr_apply_always_sensitive_never_extractable_attrs(
     }
     new_attr = NULL;
 
+    /*
+     * Also apply CKA_IBM_PROTKEY_NEVER_EXTRACTABLE if attribute
+     * CKA_IBM_PROTKEY_EXTRACTABLE is available in the key
+     */
+    rc = template_attribute_get_bool(key_obj->template,
+                                     CKA_IBM_PROTKEY_EXTRACTABLE, &flag);
+    if (rc == CKR_TEMPLATE_INCOMPLETE) {
+        rc = CKR_OK;
+        goto error;
+    }
+    if (rc != CKR_OK) {
+        TRACE_ERROR("CKA_IBM_PROTKEY_EXTRACTABLE in key object template is "
+                    "invalid.\n");
+        goto error;
+    }
+
+    rc = build_attribute(CKA_IBM_PROTKEY_NEVER_EXTRACTABLE, &ck_true,
+                         sizeof(CK_BBOOL), &new_attr);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("build_attribute failed\n");
+        goto error;
+    }
+    if (flag == TRUE)
+        *(CK_BBOOL *)new_attr->pValue = FALSE;
+
+    rc = template_update_attribute(key_obj->template, new_attr);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("template_update_attribute failed.\n");
+        goto error;
+    }
+    new_attr = NULL;
+
 error:
     if (new_attr != NULL)
         free(new_attr);
@@ -102,13 +134,105 @@ error:
     return rc;
 }
 
+CK_RV key_mgr_derive_always_sensitive_never_extractable_attrs(
+                                                    STDLL_TokData_t *tokdata,
+                                                    OBJECT *base_key_obj,
+                                                    OBJECT *derived_key_obj)
+{
+    CK_ATTRIBUTE *always_sens_attr = NULL, *never_extract_attr = NULL;
+    CK_BBOOL flag;
+    CK_RV rc;
+
+    UNUSED(tokdata);
+
+    /*
+     * If base key has ALWAYS_SENSITIVE = FALSE, then new key does too
+     * otherwise, the value of CKA_ALWAYS_SENSITIVE = CKA_SENSITIVE
+     */
+    rc = template_attribute_get_bool(base_key_obj->template,
+                                     CKA_ALWAYS_SENSITIVE, &flag);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Could not find CKA_ALWAYS_SENSITIVE in the template\n");
+        goto end;
+    }
+
+    if (flag == TRUE) {
+        rc = template_attribute_get_bool(derived_key_obj->template,
+                                         CKA_SENSITIVE, &flag);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_SENSITIVE in the template\n");
+            goto end;
+        }
+    }
+
+    rc = build_attribute(CKA_ALWAYS_SENSITIVE, &flag, sizeof(CK_BBOOL),
+                         &always_sens_attr);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Failed to build CKA_ALWAYS_SENSITIVE attribute.\n");
+        goto end;
+    }
+
+    /*
+     * If base key has NEVER_EXTRACTABLE = FALSE, the new key does too
+     * otherwise, the value of CKA_NEVER_EXTRACTABLE = !CKA_EXTRACTABLE
+     */
+    rc = template_attribute_get_bool(base_key_obj->template,
+                                     CKA_NEVER_EXTRACTABLE, &flag);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Could not find CKA_NEVER_EXTRACTABLE in the template.\n");
+        goto end;
+    }
+
+    if (flag == TRUE) {
+        rc = template_attribute_get_bool(derived_key_obj->template,
+                                         CKA_EXTRACTABLE, &flag);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("Could not find CKA_EXTRACTABLE in the template.\n");
+            goto end;
+        }
+
+        flag = !flag;
+    }
+
+    rc = build_attribute(CKA_NEVER_EXTRACTABLE, &flag, sizeof(CK_BBOOL),
+                         &never_extract_attr);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Failed to build CKA_NEVER_EXTRACTABLE attribute.\n");
+        goto end;
+    }
+
+    rc = template_update_attribute(derived_key_obj->template, always_sens_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto end;
+    }
+    always_sens_attr = NULL;
+
+    rc = template_update_attribute(derived_key_obj->template,
+                                   never_extract_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto end;
+    }
+    never_extract_attr = NULL;
+
+end:
+    if (always_sens_attr != NULL)
+        free(always_sens_attr);
+    if (never_extract_attr != NULL)
+        free(never_extract_attr);
+
+    return rc;
+}
+
 //
 //
 CK_RV key_mgr_generate_key(STDLL_TokData_t *tokdata,
                            SESSION *sess,
                            CK_MECHANISM *mech,
                            CK_ATTRIBUTE *pTemplate,
-                           CK_ULONG ulCount, CK_OBJECT_HANDLE *handle)
+                           CK_ULONG ulCount, CK_OBJECT_HANDLE *handle,
+                           CK_BBOOL count_statistics)
 {
     OBJECT *key_obj = NULL;
     CK_ATTRIBUTE *new_attr = NULL;
@@ -304,7 +428,7 @@ CK_RV key_mgr_generate_key(STDLL_TokData_t *tokdata,
         goto error;
     }
 
-    if (rc == CKR_OK)
+    if (count_statistics == TRUE && rc == CKR_OK)
         INC_COUNTER(tokdata, sess, mech, key_obj, POLICY_STRENGTH_IDX_0);
 
     return rc;
@@ -331,7 +455,8 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
                                 CK_ATTRIBUTE *priv_tmpl,
                                 CK_ULONG priv_count,
                                 CK_OBJECT_HANDLE *publ_key_handle,
-                                CK_OBJECT_HANDLE *priv_key_handle)
+                                CK_OBJECT_HANDLE *priv_key_handle,
+                                CK_BBOOL count_statistics)
 {
     OBJECT *publ_key_obj = NULL;
     OBJECT *priv_key_obj = NULL;
@@ -443,6 +568,13 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
         break;
 #endif
 /* End  code contributed by Corrent corp. */
+    case CKM_IBM_DILITHIUM:
+        if (subclass != 0 && subclass != CKK_IBM_PQC_DILITHIUM) {
+            TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+        subclass = CKK_IBM_DILITHIUM;
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -494,6 +626,10 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
         break;
 #endif
 /* End code contributed by Corrent corp. */
+    case CKM_IBM_DILITHIUM:
+        rc = ckm_ibm_dilithium_key_pair_gen(tokdata, publ_key_obj->template,
+                                            priv_key_obj->template);
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         rc = CKR_MECHANISM_INVALID;
@@ -617,7 +753,7 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
         goto error;
     }
 
-    if (rc == CKR_OK)
+    if (count_statistics == TRUE && rc == CKR_OK)
         INC_COUNTER(tokdata, sess, mech, priv_key_obj, POLICY_STRENGTH_IDX_0);
 
     return rc;
@@ -647,7 +783,8 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
                        CK_MECHANISM *mech,
                        CK_OBJECT_HANDLE h_wrapping_key,
                        CK_OBJECT_HANDLE h_key,
-                       CK_BYTE *wrapped_key, CK_ULONG *wrapped_key_len)
+                       CK_BYTE *wrapped_key, CK_ULONG *wrapped_key_len,
+                       CK_BBOOL count_statistics)
 {
     ENCR_DECR_CONTEXT *ctx = NULL;
     OBJECT *wrapping_key_obj = NULL;
@@ -794,6 +931,13 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
     case CKM_AES_CFB64:
     case CKM_AES_CFB128:
     case CKM_AES_XTS:
+    case CKM_AES_GCM:
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_KWP:
+    case CKM_AES_KEY_WRAP_PKCS7:
+    case CKM_RSA_AES_KEY_WRAP:
+    case CKM_ECDH_AES_KEY_WRAP:
         if ((class != CKO_SECRET_KEY) && (class != CKO_PRIVATE_KEY)) {
             TRACE_ERROR
                 ("Specified mechanism only wraps secret & private keys.\n");
@@ -849,7 +993,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
 
     switch (keytype) {
     case CKK_DES:
-        rc = des_wrap_get_data(key_obj->template, length_only, &data,
+        rc = des_wrap_get_data(key_obj->template, FALSE, &data,
                                &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("des_wrap_get_data failed.\n");
@@ -857,7 +1001,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         }
         break;
     case CKK_DES3:
-        rc = des3_wrap_get_data(key_obj->template, length_only, &data,
+        rc = des3_wrap_get_data(key_obj->template, FALSE, &data,
                                 &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("des3_wrap_get_data failed.\n");
@@ -865,7 +1009,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         }
         break;
     case CKK_RSA:
-        rc = rsa_priv_wrap_get_data(key_obj->template, length_only, &data,
+        rc = rsa_priv_wrap_get_data(key_obj->template, FALSE, &data,
                                     &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("rsa_priv_wrap_get_data failed.\n");
@@ -874,7 +1018,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         break;
 #if !(NODSA)
     case CKK_DSA:
-        rc = dsa_priv_wrap_get_data(key_obj->template, length_only, &data,
+        rc = dsa_priv_wrap_get_data(key_obj->template, FALSE, &data,
                                     &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("dsa_priv_wrap_get_data failed.\n");
@@ -884,7 +1028,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
 #endif
 #if !(NODH)
     case CKK_DH:
-        rc = dh_priv_wrap_get_data(key_obj->template, length_only, &data,
+        rc = dh_priv_wrap_get_data(key_obj->template, FALSE, &data,
                                    &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("dh_priv_wrap_get_data failed.\n");
@@ -893,7 +1037,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         break;
 #endif
     case CKK_GENERIC_SECRET:
-        rc = generic_secret_wrap_get_data(key_obj->template, length_only,
+        rc = generic_secret_wrap_get_data(key_obj->template, FALSE,
                                           &data, &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("generic_secret_wrap_get_data failed.\n");
@@ -902,7 +1046,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         break;
     case CKK_AES:
     case CKK_AES_XTS:
-        rc = aes_wrap_get_data(key_obj->template, length_only, &data,
+        rc = aes_wrap_get_data(key_obj->template, FALSE, &data,
                                &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("aes_wrap_get_data failed.\n");
@@ -910,11 +1054,26 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         }
         break;
     case CKK_EC:
-        rc = ecdsa_priv_wrap_get_data(key_obj->template, length_only, &data,
+        rc = ecdsa_priv_wrap_get_data(key_obj->template, FALSE, &data,
                                       &data_len);
         if (rc != CKR_OK) {
-            TRACE_DEVEL("ecdsa_priv_wrap_get_data failed with rc=%s.\n",
-                        ock_err(rc));
+            TRACE_DEVEL("ecdsa_priv_wrap_get_data failed.\n");
+            goto done;
+        }
+        break;
+    case CKK_IBM_PQC_DILITHIUM:
+        rc = ibm_dilithium_priv_wrap_get_data(key_obj->template, FALSE,
+                                              &data, &data_len);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("ibm_dilithium_priv_wrap_get_data failed.\n");
+            goto done;
+        }
+        break;
+    case CKK_IBM_PQC_KYBER:
+        rc = ibm_kyber_priv_wrap_get_data(key_obj->template, FALSE,
+                                          &data, &data_len);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("ibm_kyber_priv_wrap_get_data failed.\n");
             goto done;
         }
         break;
@@ -931,7 +1090,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
     case CKM_DES_CBC:
     case CKM_DES3_ECB:
     case CKM_DES3_CBC:
-        rc = ckm_des_wrap_format(tokdata, length_only, &data, &data_len);
+        rc = ckm_des_wrap_format(tokdata, FALSE, &data, &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("ckm_des_wrap_format failed.\n");
             if (data) {
@@ -948,7 +1107,11 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
     case CKM_AES_CFB8:
     case CKM_AES_CFB64:
     case CKM_AES_CFB128:
-        rc = ckm_aes_wrap_format(tokdata, length_only, &data, &data_len);
+    case CKM_AES_KEY_WRAP:
+        rc = ckm_aes_wrap_format(tokdata, FALSE,
+                                 mech->mechanism == CKM_AES_KEY_WRAP ?
+                                     AES_KEY_WRAP_BLOCK_SIZE : AES_BLOCK_SIZE,
+                                 &data, &data_len);
         if (rc != CKR_OK) {
             TRACE_DEVEL("ckm_aes_wrap_format failed.\n");
             if (data) {
@@ -958,6 +1121,11 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
             goto done;
         }
         break;
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_KWP:
+    case CKM_AES_KEY_WRAP_PKCS7:
+    case CKM_RSA_AES_KEY_WRAP:
+    case CKM_ECDH_AES_KEY_WRAP:
     case CKM_DES_CBC_PAD:
     case CKM_DES3_CBC_PAD:
     case CKM_AES_CBC_PAD:
@@ -968,6 +1136,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
     case CKM_RSA_PKCS:
     case CKM_RSA_X_509:
     case CKM_AES_XTS:
+    case CKM_AES_GCM:
         break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
@@ -1005,8 +1174,8 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         }
         goto done;
     }
-    // do the encryption and clean up.  at this point, 'value' may or may not
-    // be NULL depending on 'length_only'
+    // do the encryption and clean up.  at this point, 'wrapped_key' may or may
+    // not be NULL depending on 'length_only'
     //
     rc = encr_mgr_encrypt(tokdata, sess, length_only,
                           ctx, data, data_len, wrapped_key, wrapped_key_len);
@@ -1018,7 +1187,7 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
     free(ctx);
 
 done:
-    if (rc == CKR_OK)
+    if (count_statistics == TRUE && rc == CKR_OK)
         INC_COUNTER(tokdata, sess, mech, wrapping_key_obj,
                     POLICY_STRENGTH_IDX_0);
 
@@ -1045,7 +1214,8 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
                          CK_BYTE *wrapped_key,
                          CK_ULONG wrapped_key_len,
                          CK_OBJECT_HANDLE h_unwrapping_key,
-                         CK_OBJECT_HANDLE *h_unwrapped_key)
+                         CK_OBJECT_HANDLE *h_unwrapped_key,
+                         CK_BBOOL count_statistics)
 {
     ENCR_DECR_CONTEXT *ctx = NULL;
     OBJECT *key_obj = NULL, *unwrapping_key_obj = NULL;
@@ -1055,6 +1225,7 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
     CK_BBOOL fromend, not_opaque = FALSE, flag;
     CK_ATTRIBUTE *new_attrs = NULL;
     CK_ULONG new_attr_count = 0;
+    CK_MECHANISM *statistics_mech = mech;
     CK_RV rc;
 
     if (!sess || !wrapped_key || !h_unwrapped_key) {
@@ -1150,6 +1321,13 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
     case CKM_DES3_CBC_PAD:
     case CKM_AES_CBC_PAD:
     case CKM_AES_XTS:
+    case CKM_AES_GCM:
+    case CKM_AES_KEY_WRAP:
+    case CKM_AES_KEY_WRAP_PAD:
+    case CKM_AES_KEY_WRAP_KWP:
+    case CKM_AES_KEY_WRAP_PKCS7:
+    case CKM_RSA_AES_KEY_WRAP:
+    case CKM_ECDH_AES_KEY_WRAP:
         if ((keyclass != CKO_SECRET_KEY) && (keyclass != CKO_PRIVATE_KEY)) {
             TRACE_ERROR("Specified mech unwraps secret & private keys only.\n");
             rc = CKR_ARGUMENTS_BAD;
@@ -1192,6 +1370,8 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
         case CKM_AES_CFB64:
         case CKM_AES_CFB128:
         case CKM_AES_XTS:
+        case CKM_AES_GCM:
+        case CKM_AES_KEY_WRAP:
             if (keytype != CKK_AES && keytype != CKK_AES_XTS &&
                 keytype != CKK_GENERIC_SECRET) {
                 TRACE_ERROR("The key type does not allow CKA_VALUE_LEN to be "
@@ -1246,7 +1426,7 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
 
     /* Policy already checked */
     rc = decr_mgr_init(tokdata, sess, ctx, OP_UNWRAP, mech, h_unwrapping_key,
-                       FALSE);
+                       FALSE, FALSE);
     if (rc != CKR_OK)
         goto done;
 
@@ -1269,11 +1449,6 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
     rc = decr_mgr_decrypt(tokdata, sess,
                           FALSE,
                           ctx, wrapped_key, wrapped_key_len, data, &data_len);
-
-    decr_mgr_cleanup(tokdata, sess, ctx);
-    free(ctx);
-    ctx = NULL;
-
     if (rc != CKR_OK) {
         if (rc == CKR_ENCRYPTED_DATA_LEN_RANGE)
             rc = CKR_WRAPPED_KEY_LEN_RANGE;
@@ -1291,6 +1466,18 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
     else
         fromend = FALSE;
 
+    if (mech->mechanism == CKM_RSA_AES_KEY_WRAP) {
+        /*
+         * For CKM_RSA_AES_KEY_WRAP, the mechanism parameter copy in ctx has
+         * been updated to hold the actual size of the temp. AES key. Use this
+         * for statistics counting, not the user supplied mechanism parameter,
+         * which might contain an invalid/incorrect AES key size.
+         * Ctx is cleaned up and freed only after statistics counting, so the
+         * copied (and updated) mechanism parameter is still available.
+         */
+        statistics_mech = &ctx->mech;
+    }
+
     // extract the key type from the PrivateKeyInfo::AlgorithmIndicator
     if (keyclass == CKO_PRIVATE_KEY) {
         rc = key_mgr_get_private_key_type(data, data_len, &priv_keytype);
@@ -1332,6 +1519,18 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
         goto done;
     }
 
+    /* Give the token a chance to check if it supports the unwrapped key */
+    if (token_specific.t_set_attrs_for_new_object != NULL) {
+        rc = token_specific.t_set_attrs_for_new_object(tokdata, keyclass,
+                                                       MODE_UNWRAPPED,
+                                                       key_obj->template);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("token_specific.t_set_attrs_for_new_object failed with "
+                        "rc=%lx\n", rc);
+            goto done;
+        }
+    }
+
 final:
     // at this point, the key should be fully constructed...assign
     // an object handle and store the key
@@ -1343,8 +1542,8 @@ final:
     }
 
 done:
-    if (rc == CKR_OK)
-        INC_COUNTER(tokdata, sess, mech, unwrapping_key_obj,
+    if (count_statistics == TRUE && rc == CKR_OK)
+        INC_COUNTER(tokdata, sess, statistics_mech, unwrapping_key_obj,
                     POLICY_STRENGTH_IDX_0);
 
     if (rc != CKR_OK && key_obj)
@@ -1448,7 +1647,8 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
                          CK_MECHANISM *mech,
                          CK_OBJECT_HANDLE base_key,
                          CK_OBJECT_HANDLE *derived_key,
-                         CK_ATTRIBUTE *pTemplate, CK_ULONG ulCount)
+                         CK_ATTRIBUTE *pTemplate, CK_ULONG ulCount,
+                         CK_BBOOL count_statistics)
 {
     OBJECT *base_key_obj = NULL;
     CK_ATTRIBUTE *new_attrs = NULL;
@@ -1518,11 +1718,13 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
             break;
         }
         rc = ssl3_master_key_derive(tokdata, sess, mech, base_key_obj,
-                                    new_attrs, new_attr_count, derived_key);
+                                    new_attrs, new_attr_count, derived_key,
+                                    count_statistics);
         break;
     case CKM_SSL3_KEY_AND_MAC_DERIVE:
         rc = ssl3_key_and_mac_derive(tokdata, sess, mech, base_key_obj,
-                                     new_attrs, new_attr_count);
+                                     new_attrs, new_attr_count,
+                                     count_statistics);
         break;
 /* Begin code contributed by Corrent corp. */
 #ifndef NODH
@@ -1533,7 +1735,8 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
             break;
         }
         rc = dh_pkcs_derive(tokdata, sess, mech, base_key_obj,
-                            new_attrs, new_attr_count, derived_key);
+                            new_attrs, new_attr_count, derived_key,
+                            count_statistics);
         break;
 #endif
 /* End code contributed by Corrent corp. */
@@ -1544,6 +1747,34 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
             break;
         }
         rc = ecdh_pkcs_derive(tokdata, sess, mech, base_key_obj, new_attrs,
+                              new_attr_count, derived_key,
+                              count_statistics);
+        break;
+    case CKM_SHA1_KEY_DERIVATION:
+    case CKM_SHA224_KEY_DERIVATION:
+    case CKM_SHA256_KEY_DERIVATION:
+    case CKM_SHA384_KEY_DERIVATION:
+    case CKM_SHA512_KEY_DERIVATION:
+    case CKM_SHA3_224_KEY_DERIVATION:
+    case CKM_SHA3_256_KEY_DERIVATION:
+    case CKM_SHA3_384_KEY_DERIVATION:
+    case CKM_SHA3_512_KEY_DERIVATION:
+        if (!derived_key) {
+            TRACE_ERROR("%s received bad argument(s)\n", __func__);
+            rc = CKR_FUNCTION_FAILED;
+            break;
+        }
+        rc = ckm_sha_derive(tokdata, sess, mech, base_key_obj, new_attrs,
+                            new_attr_count, derived_key, count_statistics);
+        break;
+    case CKM_SHAKE_128_KEY_DERIVATION:
+    case CKM_SHAKE_256_KEY_DERIVATION:
+        if (!derived_key) {
+            TRACE_ERROR("%s received bad argument(s)\n", __func__);
+            rc = CKR_FUNCTION_FAILED;
+            break;
+        }
+        rc = ckm_shake_derive(tokdata, sess, mech, base_key_obj, new_attrs,
                               new_attr_count, derived_key);
         break;
     default:
