@@ -948,15 +948,12 @@ CK_RV des3_ecb_encrypt_final(STDLL_TokData_t *tokdata,
 
     UNUSED(tokdata);
     UNUSED(out_data);
+    UNUSED(length_only);
 
     if (!sess || !ctx || !out_data_len) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
-    // satisfy the compiler
-    //
-    if (length_only)
-        context = NULL;
 
     context = (DES_CONTEXT *) ctx->context;
 
@@ -986,15 +983,12 @@ CK_RV des3_ecb_decrypt_final(STDLL_TokData_t *tokdata,
 
     UNUSED(tokdata);
     UNUSED(out_data);
+    UNUSED(length_only);
 
     if (!sess || !ctx || !out_data_len) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
-    // satisfy the compiler
-    //
-    if (length_only)
-        context = NULL;
 
     context = (DES_CONTEXT *) ctx->context;
 
@@ -1025,15 +1019,12 @@ CK_RV des3_cbc_encrypt_final(STDLL_TokData_t *tokdata,
 
     UNUSED(tokdata);
     UNUSED(out_data);
+    UNUSED(length_only);
 
     if (!sess || !ctx || !out_data_len) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
-    // satisfy the compiler
-    //
-    if (length_only)
-        context = NULL;
 
     context = (DES_CONTEXT *) ctx->context;
 
@@ -1064,15 +1055,12 @@ CK_RV des3_cbc_decrypt_final(STDLL_TokData_t *tokdata,
 
     UNUSED(tokdata);
     UNUSED(out_data);
+    UNUSED(length_only);
 
     if (!sess || !ctx || !out_data_len) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
-    // satisfy the compiler
-    //
-    if (length_only)
-        context = NULL;
 
     context = (DES_CONTEXT *) ctx->context;
 
