@@ -46,14 +46,16 @@ CK_RV CreateXProcLock(char *tokname, STDLL_TokData_t *tokdata)
     char lockdir[PATH_MAX];
     struct group *grp;
     struct stat statbuf;
-    mode_t mode = (S_IRUSR | S_IRGRP);
     int ret = -1;
-    char *toklockname;
+    char *toklockname, *group;
+
+    group = tokdata->tokgroup;
+    if (group == NULL || group[0] == '\0')
+        group = PKCS_GROUP;
 
     if (tokdata->spinxplfd == -1) {
-
         if (token_specific.t_creatlock != NULL) {
-            tokdata->spinxplfd = token_specific.t_creatlock();
+            tokdata->spinxplfd = token_specific.t_creatlock(tokdata);
             if (tokdata->spinxplfd != -1)
                 return CKR_OK;
             else
@@ -61,10 +63,11 @@ CK_RV CreateXProcLock(char *tokname, STDLL_TokData_t *tokdata)
         }
 
         toklockname = (strlen(tokname) > 0) ? tokname : SUB_DIR;
-            
 
-        /** create lock subdir for each token if it doesn't exist.
-         *  The root directory should be created in slotmgr daemon **/
+        /*
+         * create lock subdir for each token if it doesn't exist.
+         * The root directory should be created in slotmgr daemon
+         */
         if (ock_snprintf(lockdir, PATH_MAX, "%s/%s",
                          LOCKDIR_PATH, toklockname) != 0) {
             OCK_SYSLOG(LOG_ERR, "lock directory path too long\n");
@@ -72,35 +75,53 @@ CK_RV CreateXProcLock(char *tokname, STDLL_TokData_t *tokdata)
             goto err;
         }
 
+        grp = getgrnam(group);
+        if (grp == NULL) {
+            OCK_SYSLOG(LOG_ERR, "getgrname(%s): %s\n", group, strerror(errno));
+            TRACE_ERROR("getgrname(%s): %s\n", group, strerror(errno));
+            goto err;
+        }
+
         ret = stat(lockdir, &statbuf);
         if (ret != 0 && errno == ENOENT) {
             /* dir does not exist, try to create it */
             ret = mkdir(lockdir, S_IRWXU | S_IRWXG);
             if (ret != 0) {
-                OCK_SYSLOG(LOG_ERR,
-                           "Directory(%s) missing: %s\n",
+                OCK_SYSLOG(LOG_ERR, "Directory(%s) missing: %s\n",
                            lockdir, strerror(errno));
+                TRACE_ERROR("Directory(%s) missing: %s\n",
+                            lockdir, strerror(errno));
                 goto err;
             }
-            grp = getgrnam(PKCS_GROUP);
-            if (grp == NULL) {
-                fprintf(stderr, "getgrname(%s): %s", PKCS_GROUP,
-                        strerror(errno));
-                goto err;
-            }
-            /* set ownership to euid, and pkcs11 group */
+
+            /* set ownership to euid, and token group */
             if (chown(lockdir, geteuid(), grp->gr_gid) != 0) {
-                fprintf(stderr, "Failed to set owner:group \
-                        ownership on %s directory", lockdir);
+                OCK_SYSLOG(LOG_ERR, "Failed to set owner:group ownership on "
+                           "'%s' directory\n", lockdir);
+                TRACE_ERROR("Failed to set owner:group ownership on '%s' "
+                            "directory\n", lockdir);
                 goto err;
             }
-            /* mkdir does not set group permission right, so
-             ** trying explictly here again */
+            /* mkdir does not set group permission right, set explicitly here */
             if (chmod(lockdir, S_IRWXU | S_IRWXG) != 0) {
-                fprintf(stderr, "Failed to change \
-                        permissions on %s directory", lockdir);
+                OCK_SYSLOG(LOG_ERR, "Failed to change permissions on '%s' "
+                           "directory\n", lockdir);
+                TRACE_ERROR("Failed to change permissions on '%s' directory\n",
+                            lockdir);
                 goto err;
             }
+        } else if (ret != 0) {
+            OCK_SYSLOG(LOG_ERR, "Could not stat directory '%s': %s\n", lockdir,
+                       strerror(errno));
+            TRACE_ERROR("Could not stat directory '%s': %s\n", lockdir,
+                        strerror(errno));
+            goto err;
+        } else if (statbuf.st_gid != grp->gr_gid) {
+            OCK_SYSLOG(LOG_ERR, "Directory '%s' is not owned by token group "
+                       "'%s'\n", lockdir, group);
+            TRACE_ERROR("Directory '%s' is not owned by token group '%s'\n",
+                        lockdir, group);
+            goto err;
         }
 
         /* create user lock file */
@@ -109,37 +130,34 @@ CK_RV CreateXProcLock(char *tokname, STDLL_TokData_t *tokdata)
             OCK_SYSLOG(LOG_ERR, "lock file path too long\n");
             TRACE_ERROR("lock file path too long\n");
             goto err;
-        }        
+        }
 
         if (stat(lockfile, &statbuf) == 0) {
-            tokdata->spinxplfd = open(lockfile, O_RDONLY, mode);
+            tokdata->spinxplfd = open(lockfile, OPEN_MODE);
         } else {
-            tokdata->spinxplfd = open(lockfile, O_CREAT | O_RDONLY, mode);
+            tokdata->spinxplfd = open(lockfile, O_CREAT | OPEN_MODE, MODE_BITS);
             if (tokdata->spinxplfd != -1) {
-                /* umask may prevent correct mode,so set it. */
-                if (fchmod(tokdata->spinxplfd, mode) == -1) {
+                /* umask may prevent correct mode, so set it. */
+                if (fchmod(tokdata->spinxplfd, MODE_BITS) == -1) {
                     OCK_SYSLOG(LOG_ERR, "fchmod(%s): %s\n",
                                lockfile, strerror(errno));
+                    TRACE_ERROR("fchmod(%s): %s\n", lockfile, strerror(errno));
                     goto err;
                 }
 
-                grp = getgrnam(PKCS_GROUP);
-                if (grp != NULL) {
-                    if (fchown(tokdata->spinxplfd, -1, grp->gr_gid) == -1) {
-                        OCK_SYSLOG(LOG_ERR,
-                                   "fchown(%s): %s\n",
-                                   lockfile, strerror(errno));
-                        goto err;
-                    }
-                } else {
-                    OCK_SYSLOG(LOG_ERR, "getgrnam(): %s\n", strerror(errno));
+                if (fchown(tokdata->spinxplfd, -1, grp->gr_gid) == -1) {
+                    OCK_SYSLOG(LOG_ERR, "fchown(%s): %s\n",
+                               lockfile, strerror(errno));
+                    TRACE_ERROR("fchown(%s): %s\n", lockfile,
+                                strerror(errno));
                     goto err;
                 }
             }
         }
         if (tokdata->spinxplfd == -1) {
             OCK_SYSLOG(LOG_ERR, "open(%s): %s\n", lockfile, strerror(errno));
-            return CKR_FUNCTION_FAILED;
+            TRACE_ERROR("open(%s): %s\n", lockfile, strerror(errno));
+            goto err;
         }
     }
 
@@ -299,7 +317,8 @@ void init_tokenInfo(TOKEN_DATA *nv_token_data)
     //
 
     token_info->flags = CKF_RNG | CKF_LOGIN_REQUIRED | CKF_CLOCK_ON_TOKEN |
-        CKF_SO_PIN_TO_BE_CHANGED | CKF_USER_PIN_TO_BE_CHANGED;
+        CKF_SO_PIN_TO_BE_CHANGED | CKF_USER_PIN_TO_BE_CHANGED |
+        CKF_DUAL_CRYPTO_OPERATIONS;
 
     // For the release, we made these
     // values as CK_UNAVAILABLE_INFORMATION or CK_EFFECTIVELY_INFINITE
@@ -598,7 +617,8 @@ CK_RV attach_shm(STDLL_TokData_t *tokdata, CK_SLOT_ID slot_id)
         rc = CKR_FUNCTION_FAILED;
         goto err;
     }
-    ret = sm_open(buf, 0660, (void **) shm, sizeof(**shm), 0);
+    ret = sm_open(buf, 0660, (void **) shm, sizeof(**shm), 0,
+                  tokdata->tokgroup);
     if (ret < 0) {
         TRACE_DEVEL("sm_open failed.\n");
         rc = CKR_FUNCTION_FAILED;
@@ -670,24 +690,28 @@ CK_RV compute_sha(STDLL_TokData_t *tokdata, CK_BYTE *data, CK_ULONG len,
         md = EVP_sha512();
         break;
 #ifdef NID_sha3_224
+    case CKM_SHA3_224:
     case CKM_IBM_SHA3_224:
         hash_len = SHA3_224_HASH_SIZE;
         md = EVP_sha3_224();
         break;
 #endif
 #ifdef NID_sha3_256
+    case CKM_SHA3_256:
     case CKM_IBM_SHA3_256:
         hash_len = SHA3_256_HASH_SIZE;
         md = EVP_sha3_256();
         break;
 #endif
 #ifdef NID_sha3_384
+    case CKM_SHA3_384:
     case CKM_IBM_SHA3_384:
         hash_len = SHA3_384_HASH_SIZE;
         md = EVP_sha3_384();
         break;
 #endif
 #ifdef NID_sha3_512
+    case CKM_SHA3_512:
     case CKM_IBM_SHA3_512:
         hash_len = SHA3_512_HASH_SIZE;
         md = EVP_sha3_512();
@@ -867,6 +891,12 @@ CK_RV pkcs_get_keytype(CK_ATTRIBUTE *attrs, CK_ULONG attrs_len,
     case CKM_SHA256_KEY_DERIVATION:
     case CKM_SHA384_KEY_DERIVATION:
     case CKM_SHA512_KEY_DERIVATION:
+    case CKM_SHA3_224_KEY_DERIVATION:
+    case CKM_SHA3_256_KEY_DERIVATION:
+    case CKM_SHA3_384_KEY_DERIVATION:
+    case CKM_SHA3_512_KEY_DERIVATION:
+    case CKM_SHAKE_128_KEY_DERIVATION:
+    case CKM_SHAKE_256_KEY_DERIVATION:
         *type = CKK_GENERIC_SECRET;
         break;
     case CKM_RSA_PKCS_KEY_PAIR_GEN:
@@ -949,7 +979,9 @@ CK_RV init_hsm_mk_change_lock(STDLL_TokData_t *tokdata)
     /*
      * Request the API layer to lock against HSM-MK-change state changes.
      * Set lock kind PTHREAD_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP to avoid
-     * writer starvation. Otherwise in a multi-threaded OCK application
+     * writer starvation - but only on Linux. This is a non-portable GNU
+     * extension, with no POSIX mechanism to achieve a similar functionality.
+     * Otherwise in a multi-threaded OCK application
      * with a heavy crypto workload, the event thread would never get the
      * HSM-MK-change lock as writer.
      */
@@ -960,6 +992,8 @@ CK_RV init_hsm_mk_change_lock(STDLL_TokData_t *tokdata)
         return CKR_CANT_LOCK;
     }
 
+#if !defined(_AIX)
+#if _XOPEN_SOURCE >= 500 || _POSIX_C_SOURCE >= 200809L
     if (pthread_rwlockattr_setkind_np(&attr,
                   PTHREAD_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP) != 0) {
         TRACE_ERROR("pthread_rwlockattr_setkind_np failed\n");
@@ -968,6 +1002,8 @@ CK_RV init_hsm_mk_change_lock(STDLL_TokData_t *tokdata)
         pthread_rwlockattr_destroy(&attr);
         return CKR_CANT_LOCK;
     }
+#endif /* _XOPEN_SOURCE >= 500 || _POSIX_C_SOURCE >= 200809L */
+#endif /* _AIX */
 
     if (pthread_rwlock_init(&tokdata->hsm_mk_change_rwlock, &attr) != 0) {
         TRACE_ERROR("pthread_rwlock_init failed\n");
