@@ -19,7 +19,10 @@
 #include "ep11_func.h"
 #include "configuration.h"
 
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wstrict-prototypes"
 #include <ica_api.h>
+#pragma GCC diagnostic pop
 
 #include <pthread.h>
 
@@ -43,6 +46,19 @@
 #define MAX_SIGN_STATE_BYTES        12288
 #define MAX_APQN                    256
 #define EP11_BLOB_WKID_OFFSET       32
+#define EP11_BLOB_BOOL_ATTR_OFFSET  52 /* second 32 bits of bool attr field */
+#define EP11_BLOB_VERSION_OFFSET    64
+#define EP11_BLOB_VERSION           0x1234
+
+typedef struct {
+    uint32_t version  : 4,
+             reserved : 4,
+             num_int_attrs : 8,
+             num_varlen_attrs : 4,
+             wordcount_varlen_attrs : 12;
+} __attribute__ ((packed)) blob_attr_header_t;
+
+#define EP11_BLOB_ATTR_HDR_VERSION  1
 
 typedef struct cp_mech_config {
     CK_MECHANISM_TYPE mech;
@@ -241,6 +257,8 @@ typedef struct {
 #define PKEY_MODE_DISABLED          0
 #define PKEY_MODE_DEFAULT           1
 #define PKEY_MODE_ENABLE4NONEXTR    2
+#define PKEY_MODE_ENABLE4EXTR       3
+#define PKEY_MODE_ENABLE4ALL        4
 
 #define PQC_BYTE_NO(idx)            (((idx) - 1) / 8)
 #define PQC_BIT_IN_BYTE(idx)        (((idx - 1)) % 8)
@@ -278,6 +296,7 @@ typedef struct {
     int fips_session_mode;
     int optimize_single_ops;
     int pkey_mode;
+    volatile int pkey_combined_extract_supported;
     volatile int pkey_wrap_supported;
     int pkey_wrap_support_checked;
     char pkey_mk_vp[PKEY_MK_VP_LENGTH];
@@ -324,6 +343,8 @@ typedef struct {
 #define REGEX_SUB_CARD_PATTERN  "[0-9a-fA-F]+\\.[0-9a-fA-F]+"
 #define MASK_EP11               0x04000000
 
+#define SYSFS_SEL_GUEST        "/sys/firmware/uv/prot_virt_guest"
+
 typedef struct {
     STDLL_TokData_t *tokdata;
     ep11_session_t *ep11_session;
@@ -467,13 +488,14 @@ CK_RV ep11tok_generate_key_pair(STDLL_TokData_t * tokdata, SESSION * sess,
 
 CK_RV ep11tok_check_single_mech_key(STDLL_TokData_t *tokdata, SESSION * session,
                                     CK_MECHANISM *mech, CK_OBJECT_HANDLE key,
-                                    CK_ULONG operation);
+                                    CK_ULONG operation,
+                                    CK_BBOOL *auth_required);
 
 CK_BOOL ep11tok_mech_single_only(CK_MECHANISM *mech);
 
 CK_RV ep11tok_sign_init(STDLL_TokData_t * tokdata, SESSION * session,
                         CK_MECHANISM * mech, CK_BBOOL recover_mode,
-                        CK_OBJECT_HANDLE key);
+                        CK_OBJECT_HANDLE key, CK_BBOOL checkauth);
 
 CK_RV ep11tok_sign(STDLL_TokData_t * tokdata, SESSION * session,
                    CK_BBOOL length_only, CK_BYTE * in_data,
@@ -548,7 +570,8 @@ CK_RV ep11tok_encrypt_single(STDLL_TokData_t *tokdata, SESSION *session,
                              CK_ULONG_PTR p_output_data_len);
 
 CK_RV ep11tok_decrypt_init(STDLL_TokData_t * tokdata, SESSION * session,
-                           CK_MECHANISM_PTR mech, CK_OBJECT_HANDLE key);
+                           CK_MECHANISM_PTR mech, CK_OBJECT_HANDLE key,
+                           CK_BBOOL checkauth);
 
 CK_RV ep11tok_decrypt_single(STDLL_TokData_t *tokdata, SESSION *session,
                              CK_MECHANISM *mech, CK_BBOOL length_only,
@@ -582,8 +605,8 @@ CK_BBOOL ep11tok_libica_mech_available(STDLL_TokData_t *tokdata,
 CK_RV ep11tok_copy_firmware_info(STDLL_TokData_t *tokdata,
                                  CK_TOKEN_INFO_PTR pInfo);
 
-CK_BBOOL ep11tok_pkey_usage_ok(STDLL_TokData_t *tokdata, SESSION *session,
-                               CK_OBJECT_HANDLE hkey, CK_MECHANISM *mech);
+CK_RV ep11tok_pkey_usage_ok(STDLL_TokData_t *tokdata, SESSION *session,
+                            CK_OBJECT_HANDLE hkey, CK_MECHANISM *mech);
 
 CK_RV ep11tok_set_operation_state(STDLL_TokData_t *tokdata, SESSION *session);
 
@@ -599,6 +622,9 @@ CK_RV do_LogoutExtended(XCP_LoginAlgorithm_t alg,
                         const CK_BYTE *pin, CK_ULONG pin_len,
                         const CK_BYTE *nonce, CK_ULONG nonce_len,
                         target_t target);
+CK_RV ep11tok_extract_blob_info(CK_BYTE *blob, CK_ULONG blob_len, CK_BBOOL *is_spki,
+                        CK_BYTE **wkvp, CK_BYTE** session_id,
+                        XCP_Attr_t **bool_attrs);
 
 /*
  * Macros to enclose EP11 library calls involving session bound blobs.
@@ -631,6 +657,8 @@ CK_RV do_LogoutExtended(XCP_LoginAlgorithm_t alg,
                          retry_count ++) {
 
 #define RETRY_SESSION_SINGLE_APQN_END(rc, tokdata, session)              \
+                         if ((target_info) == NULL)                      \
+                             break;                                      \
                          if (target_info->single_apqn &&                 \
                              ((rc) == CKR_IBM_TARGET_INVALID ||          \
                               ((rc) == CKR_FUNCTION_FAILED &&            \
