@@ -94,6 +94,7 @@ token_spec_t token_specific = {
     &token_specific_ec_verify,
     NULL,                       // ec_generate_keypair
     NULL,                       // ecdh_derive
+    NULL,                       // ecdh_derive_kdf
     // DH
     NULL,                       // dh_pkcs_derive,
     NULL,                       // dh_pkcs_key_pair_gen
@@ -102,6 +103,8 @@ token_spec_t token_specific = {
     &token_specific_sha,
     &token_specific_sha_update,
     &token_specific_sha_final,
+    // SHAKE derive
+    NULL,                       // shake_key_derive
     // HMAC
     NULL,                       // hmac_sign_init
     NULL,                       // hmac_sign
@@ -115,8 +118,13 @@ token_spec_t token_specific = {
     // AES
     NULL,                       // aes_key_gen,
     NULL,                       // aes_xts_key_gen
+#ifndef NO_PKEY
     &token_specific_aes_ecb,
     &token_specific_aes_cbc,
+#else
+    NULL,                       // aes_ecb
+    NULL,                       // aes_cbc
+#endif
     NULL,                       // aes_ctr
     NULL,                       // aes_gcm_init
     NULL,                       // aes_gcm
@@ -125,12 +133,22 @@ token_spec_t token_specific = {
     NULL,                       // aes_ofb
     NULL,                       // aes_cfb
     NULL,                       // aes_mac
+#ifndef NO_PKEY
     &token_specific_aes_cmac,
     &token_specific_aes_xts,    // aes_xts
+#else
+    NULL,                       // aes_cmac
+    NULL,                       // aes_xts
+#endif
+    NULL,                       // aes_key_wrap
     // DSA
     NULL,                       // dsa_generate_keypair,
     NULL,                       // dsa_sign
     NULL,                       // dsa_verify
+    // PQC
+    NULL,                       // ibm_dilithium_generate_keypair
+    NULL,                       // ibm_dilithium_sign
+    NULL,                       // ibm_dilithium_verify
     NULL,                       // get_mechanism_list
     NULL,                       // get mechanism_info
     &token_specific_object_add,
