@@ -86,14 +86,17 @@ token_spec_t token_specific = {
     NULL,                       // ec_verify
     NULL,                       // ec_generate_keypair
     NULL,                       // ecdh_derive
+    NULL,                       // ecdh_derive_kdf
     // DH
     NULL,                       // dh_pkcs_derive
     NULL,                       // dh_pkcs_key_pair_gen
     // SHA
-    NULL,                       // sha_init
-    NULL,                       // sha
-    NULL,                       // sha_update
-    NULL,                       // sha_final
+    &token_specific_sha_init,
+    &token_specific_sha,
+    &token_specific_sha_update,
+    &token_specific_sha_final,
+    // SHAKE derive
+    NULL,                       // shake_key_derive
     //HMAC
     NULL,                       // hmac_sign_init
     NULL,                       // hmac_sign
@@ -119,10 +122,15 @@ token_spec_t token_specific = {
     NULL,                       // aes_mac
     NULL,                       // aes_cmac
     NULL,                       // aes_xts
+    NULL,                       // aes_key_wrap
     // DSA
     NULL,                       // dsa_generate_keypair
     NULL,                       // dsa_sign
     NULL,                       // dsa_verify
+    // PQC
+    NULL,                       // ibm_dilithium_generate_keypair
+    NULL,                       // ibm_dilithium_sign
+    NULL,                       // ibm_dilithium_verify
     NULL,                       // get_mechanism_list
     NULL,                       // get_mechanism_info
     NULL,                       // object_add
