@@ -26,6 +26,7 @@
 
 #include <openssl/opensslv.h>
 
+#include "platform.h"
 #include "pkcs11types.h"
 #include "defs.h"
 #include "host_defs.h"
@@ -38,7 +39,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <endian.h>
 
 #include <openssl/rand.h>
 #include <openssl/dh.h>
@@ -49,6 +49,7 @@
 #if OPENSSL_VERSION_PREREQ(3, 0)
 #include <openssl/core_names.h>
 #include <openssl/param_build.h>
+#include <openssl/provider.h>
 #endif
 
 #define MAX_GENERIC_KEY_SIZE 256
@@ -73,12 +74,20 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_SHA256_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA384_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA512_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA1_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA224_RSA_PKCS_PSS, {512, 4096, CKF_SIGN|CKF_VERIFY}},
+    {CKM_SHA224_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA256_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA384_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA512_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
 #if !(NOX509)
     {CKM_RSA_X_509,
      {512, 4096, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP | CKF_SIGN |
@@ -92,6 +101,7 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
 #if !(NOMD5)
     {CKM_MD5_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
 #endif
+    {CKM_RSA_AES_KEY_WRAP, {512, 4096, CKF_WRAP | CKF_UNWRAP}},
 #if !(NODSA)
     {CKM_DSA, {512, 1024, CKF_SIGN | CKF_VERIFY}},
 #endif
@@ -120,19 +130,24 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_SHA_1, {0, 0, CKF_DIGEST}},
     {CKM_SHA_1_HMAC, {80, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA_1_HMAC_GENERAL, {80, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA1_KEY_DERIVATION, {8, 160, CKF_DERIVE}},
 #endif
     {CKM_SHA224, {0, 0, CKF_DIGEST}},
     {CKM_SHA224_HMAC, {112, 2048, CKF_SIGN|CKF_VERIFY}},
     {CKM_SHA224_HMAC_GENERAL, {112, 2048, CKF_SIGN|CKF_VERIFY}},
+    {CKM_SHA224_KEY_DERIVATION, {8, 224, CKF_DERIVE}},
     {CKM_SHA256, {0, 0, CKF_DIGEST}},
     {CKM_SHA256_HMAC, {128, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA256_HMAC_GENERAL, {128, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA256_KEY_DERIVATION, {8, 256, CKF_DERIVE}},
     {CKM_SHA384, {0, 0, CKF_DIGEST}},
     {CKM_SHA384_HMAC, {192, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA384_HMAC_GENERAL, {192, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA384_KEY_DERIVATION, {8, 384, CKF_DERIVE}},
     {CKM_SHA512, {0, 0, CKF_DIGEST}},
     {CKM_SHA512_HMAC, {256, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA512_HMAC_GENERAL, {256, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA512_KEY_DERIVATION, {8, 512, CKF_DERIVE}},
 #ifdef NID_sha512_224WithRSAEncryption
     {CKM_SHA512_224, {0, 0, CKF_DIGEST}},
     {CKM_SHA512_224_HMAC, {112, 2048, CKF_SIGN | CKF_VERIFY}},
@@ -144,18 +159,34 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_SHA512_256_HMAC_GENERAL, {128, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
 #ifdef NID_sha3_224
+    {CKM_SHA3_224, {0, 0, CKF_DIGEST}},
+    {CKM_SHA3_224_HMAC, {112, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_HMAC_GENERAL, {112, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_KEY_DERIVATION, {8, 224, CKF_DERIVE}},
     {CKM_IBM_SHA3_224, {0, 0, CKF_DIGEST}},
     {CKM_IBM_SHA3_224_HMAC, {112, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
 #ifdef NID_sha3_256
+    {CKM_SHA3_256, {0, 0, CKF_DIGEST}},
+    {CKM_SHA3_256_HMAC, {128, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_HMAC_GENERAL, {128, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_KEY_DERIVATION, {8, 256, CKF_DERIVE}},
     {CKM_IBM_SHA3_256, {0, 0, CKF_DIGEST}},
     {CKM_IBM_SHA3_256_HMAC, {128, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
 #ifdef NID_sha3_384
+    {CKM_SHA3_384, {0, 0, CKF_DIGEST}},
+    {CKM_SHA3_384_HMAC, {192, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_HMAC_GENERAL, {192, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_KEY_DERIVATION, {8, 384, CKF_DERIVE}},
     {CKM_IBM_SHA3_384, {0, 0, CKF_DIGEST}},
     {CKM_IBM_SHA3_384_HMAC, {192, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
 #ifdef NID_sha3_512
+    {CKM_SHA3_512, {0, 0, CKF_DIGEST}},
+    {CKM_SHA3_512_HMAC, {256, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_HMAC_GENERAL, {256, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_KEY_DERIVATION, {8, 512, CKF_DERIVE}},
     {CKM_IBM_SHA3_512, {0, 0, CKF_DIGEST}},
     {CKM_IBM_SHA3_512_HMAC, {256, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
@@ -169,6 +200,8 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_MD5_HMAC, {8, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_MD5_HMAC_GENERAL, {8, 2048, CKF_SIGN | CKF_VERIFY}},
 #endif
+    {CKM_SHAKE_128_KEY_DERIVATION, {8, 2048, CKF_DERIVE}},
+    {CKM_SHAKE_256_KEY_DERIVATION, {8, 2048, CKF_DERIVE}},
     {CKM_SSL3_PRE_MASTER_KEY_GEN, {48, 48, CKF_GENERATE}},
     {CKM_SSL3_MASTER_KEY_DERIVE, {48, 48, CKF_DERIVE}},
     {CKM_SSL3_KEY_AND_MAC_DERIVE, {48, 48, CKF_DERIVE}},
@@ -191,12 +224,20 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_AES_CFB8, {16, 32, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
     {CKM_AES_CFB128, {16, 32, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
 #endif
-    {CKM_AES_GCM, {16, 32, CKF_ENCRYPT | CKF_DECRYPT}},
+    {CKM_AES_GCM, {16, 32, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
     {CKM_AES_MAC, {16, 32, CKF_HW | CKF_SIGN | CKF_VERIFY}},
     {CKM_AES_MAC_GENERAL, {16, 32, CKF_HW | CKF_SIGN | CKF_VERIFY}},
     {CKM_AES_CMAC, {16, 32, CKF_SIGN | CKF_VERIFY}},
     {CKM_AES_CMAC_GENERAL, {16, 32, CKF_SIGN | CKF_VERIFY}},
     {CKM_AES_XTS, {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
+    {CKM_AES_KEY_WRAP,
+     {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
+    {CKM_AES_KEY_WRAP_PAD,
+     {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
+    {CKM_AES_KEY_WRAP_KWP,
+     {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
+    {CKM_AES_KEY_WRAP_PKCS7,
+     {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
     {CKM_GENERIC_SECRET_KEY_GEN, {80, 2048, CKF_GENERATE}},
 #if !(NO_EC)
     {CKM_EC_KEY_PAIR_GEN, {160, 521, CKF_GENERATE_KEY_PAIR |
@@ -213,43 +254,143 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
                         CKF_EC_F_P}},
     {CKM_ECDSA_SHA512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
                         CKF_EC_F_P}},
+    {CKM_ECDSA_SHA3_224, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
+                          CKF_EC_F_P}},
+    {CKM_ECDSA_SHA3_256, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
+                          CKF_EC_F_P}},
+    {CKM_ECDSA_SHA3_384, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
+                          CKF_EC_F_P}},
+    {CKM_ECDSA_SHA3_512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
+                          CKF_EC_F_P}},
     {CKM_ECDH1_DERIVE, {160, 521, CKF_DERIVE | CKF_EC_NAMEDCURVE | CKF_EC_F_P}},
+    {CKM_ECDH_AES_KEY_WRAP, {160, 521, CKF_WRAP | CKF_UNWRAP |
+                             CKF_EC_NAMEDCURVE | CKF_EC_F_P}},
+#endif
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    {CKM_IBM_DILITHIUM, {256, 256, CKF_GENERATE_KEY_PAIR |
+                                   CKF_SIGN | CKF_VERIFY}},
 #endif
 };
 
 static const CK_ULONG soft_mech_list_len =
                     (sizeof(soft_mech_list) / sizeof(MECH_LIST_ELEMENT));
 
+struct soft_private_data {
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    OSSL_PROVIDER *oqs_provider;
+#else
+    void *dummy;
+#endif
+};
+
 CK_RV token_specific_init(STDLL_TokData_t *tokdata, CK_SLOT_ID SlotNumber,
                           char *conf_name)
 {
-    UNUSED(conf_name);
+    struct soft_private_data *soft_private;
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    const struct pqc_oid *oid;
+#endif
     CK_RV rc;
 
+    UNUSED(conf_name);
+
+    TRACE_INFO("soft %s slot=%lu running\n", __func__, SlotNumber);
+
     rc = ock_generic_filter_mechanism_list(tokdata,
                                            soft_mech_list, soft_mech_list_len,
                                            &(tokdata->mech_list),
                                            &(tokdata->mech_list_len));
     if (rc != CKR_OK) {
         TRACE_ERROR("Mechanism filtering failed!  rc = 0x%lx\n", rc);
-        return rc;
+        goto error;
     }
 
-    TRACE_INFO("soft %s slot=%lu running\n", __func__, SlotNumber);
+    soft_private = calloc(1, sizeof(*soft_private));
+    if (soft_private == NULL) {
+        TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+        rc = CKR_HOST_MEMORY;
+        goto error;
+    }
+
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    /*
+     * Try to load the 'oqsprovider'. This optional provider must be installed
+     * and configured separately, it does not come with OpenSSL 3.x by default.
+     * If loading the 'oqsprovider' fails, this is not an error, it just means
+     * that the soft token does not support any quantum safe mechanisms.
+     * Also check if the 'oqsprovider' does support the Dilithium R3_44 variant.
+     * If not don't consider it loaded, because it can't support any of the
+     * Dilithium variants supported by the Soft token.
+     */
+    soft_private->oqs_provider = OSSL_PROVIDER_load(NULL, "oqsprovider");
+    if (soft_private->oqs_provider == NULL) {
+        TRACE_DEVEL("OSSL_PROVIDER_load for 'oqsprovider' failed, no quantum "
+                    "safe mechanisms are supported.\n");
+        ERR_pop_to_mark();
+    } else {
+        oid = find_pqc_by_keyform(dilithium_oids,
+                                  CK_IBM_DILITHIUM_KEYFORM_ROUND3_44);
+        if (oid == NULL || openssl_get_pqc_oid_name(oid) == NULL) {
+            OSSL_PROVIDER_unload(soft_private->oqs_provider);
+            soft_private->oqs_provider = NULL;
+            TRACE_DEVEL("The 'oqsprovider' does not support Dilithium R3_44, "
+                        "no quantum safe mechanisms are supported.\n");
+            ERR_pop_to_mark();
+        }
+    }
+#endif
+
+    tokdata->private_data = soft_private;
+
+    return CKR_OK;
+
+error:
+    token_specific_final(tokdata, FALSE);
+    return rc;
+}
+
+CK_RV token_specific_final(STDLL_TokData_t *tokdata,
+                           CK_BBOOL in_fork_initializer)
+{
+    struct soft_private_data *soft_private = tokdata->private_data;
+
+    UNUSED(in_fork_initializer);
+
+    TRACE_INFO("soft %s running\n", __func__);
+
+    if (tokdata->mech_list != NULL)
+        free(tokdata->mech_list);
+    
+    if (soft_private != NULL) {
+#if OPENSSL_VERSION_PREREQ(3, 0)
+        if (soft_private->oqs_provider != NULL)
+            OSSL_PROVIDER_unload(soft_private->oqs_provider);
+        soft_private->oqs_provider = NULL;
+#endif
+        free(soft_private);
+        tokdata->private_data = NULL;
+    }
 
     return CKR_OK;
 }
 
-CK_RV token_specific_final(STDLL_TokData_t *tokdata,
-                           CK_BBOOL token_specific_final)
+static CK_BBOOL token_specific_filter_mechanism(STDLL_TokData_t *tokdata,
+                                                CK_MECHANISM_TYPE mechanism)
 {
-    UNUSED(token_specific_final);
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    struct soft_private_data *soft_private = tokdata->private_data;
+#else
+    UNUSED(tokdata);
+#endif
 
-    TRACE_INFO("soft %s running\n", __func__);
-
-    free(tokdata->mech_list);
-    
-    return CKR_OK;
+    switch(mechanism) {
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    case CKM_IBM_DILITHIUM:
+        return soft_private->oqs_provider != NULL ? CK_TRUE : CK_FALSE;
+#endif
+    default:
+        return CK_TRUE;
+    }
 }
 
 CK_RV token_specific_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
@@ -665,11 +806,13 @@ CK_RV token_specific_aes_mac(STDLL_TokData_t *tokdata, CK_BYTE *message,
     return openssl_specific_aes_mac(tokdata, message, message_len, key, mac);
 }
 
-CK_RV token_specific_aes_cmac(STDLL_TokData_t *tokdata, CK_BYTE *message,
+CK_RV token_specific_aes_cmac(STDLL_TokData_t *tokdata, SESSION *session, CK_BYTE *message,
                               CK_ULONG message_len, OBJECT *key, CK_BYTE *mac,
                               CK_BBOOL first, CK_BBOOL last, CK_VOID_PTR *ctx)
 {
 
+    UNUSED(session);
+
     return openssl_specific_aes_cmac(tokdata, message, message_len, key, mac,
                                      first, last, ctx);
 }
@@ -688,6 +831,19 @@ CK_RV token_specific_aes_xts(STDLL_TokData_t *tokdata, SESSION *sess,
                                     tweak, encrypt, initial, final, iv);
 }
 
+CK_RV token_specific_aes_key_wrap(STDLL_TokData_t *tokdata, SESSION *sess,
+                                  CK_BYTE *in_data, CK_ULONG in_data_len,
+                                  CK_BYTE *out_data, CK_ULONG *out_data_len,
+                                  OBJECT *key_obj, CK_BYTE *iv, CK_ULONG iv_len,
+                                  CK_BBOOL encrypt, CK_BBOOL pad)
+{
+    UNUSED(sess);
+
+    return openssl_specific_aes_key_wrap(tokdata, in_data, in_data_len,
+                                         out_data, out_data_len, key_obj,
+                                         iv, iv_len, encrypt, pad);
+}
+
 /* Begin code contributed by Corrent corp. */
 #ifndef NODH
 // This computes DH shared secret, where:
@@ -788,7 +944,6 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
                                           TEMPLATE *priv_tmpl)
 {
     CK_RV rv;
-    CK_BBOOL rc;
     CK_ATTRIBUTE *prime_attr = NULL;
     CK_ATTRIBUTE *base_attr = NULL;
     CK_ATTRIBUTE *temp_attr = NULL;
@@ -948,7 +1103,7 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
 #else
     if (!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_PUB_KEY, &temp_bn)) {
         TRACE_ERROR("%s\n", ock_err(ERR_FUNCTION_FAILED));
-        rc = CKR_FUNCTION_FAILED;
+        rv = CKR_FUNCTION_FAILED;
         goto done;
     }
 #endif
@@ -957,13 +1112,13 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
     temp_byte = malloc(temp_bn_len);
     temp_bn_len = BN_bn2bin(temp_bn, temp_byte);
     // in bytes
-    rc = build_attribute(CKA_VALUE, temp_byte, temp_bn_len, &temp_attr);
-    if (rc != CKR_OK) {
+    rv = build_attribute(CKA_VALUE, temp_byte, temp_bn_len, &temp_attr);
+    if (rv != CKR_OK) {
         TRACE_DEVEL("build_attribute failed\n");
         goto done;
     }
-    rc = template_update_attribute(publ_tmpl, temp_attr);
-    if (rc != CKR_OK) {
+    rv = template_update_attribute(publ_tmpl, temp_attr);
+    if (rv != CKR_OK) {
         TRACE_ERROR("template_update_attribute failed\n");
         free(temp_attr);
         goto done;
@@ -981,7 +1136,7 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
 #else
     if (!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_PRIV_KEY, &temp_bn)) {
         TRACE_ERROR("%s\n", ock_err(ERR_FUNCTION_FAILED));
-        rc = CKR_FUNCTION_FAILED;
+        rv = CKR_FUNCTION_FAILED;
         goto done;
     }
 #endif
@@ -989,14 +1144,14 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
     temp_byte2 = malloc(temp_bn_len);
     temp_bn_len = BN_bn2bin(temp_bn, temp_byte2);
     // in bytes
-    rc = build_attribute(CKA_VALUE, temp_byte2, temp_bn_len, &temp_attr);
+    rv = build_attribute(CKA_VALUE, temp_byte2, temp_bn_len, &temp_attr);
     OPENSSL_cleanse(temp_byte2, temp_bn_len);
-    if (rc != CKR_OK) {
+    if (rv != CKR_OK) {
         TRACE_DEVEL("build_attribute failed\n");
         goto done;
     }
-    rc = template_update_attribute(priv_tmpl, temp_attr);
-    if (rc != CKR_OK) {
+    rv = template_update_attribute(priv_tmpl, temp_attr);
+    if (rv != CKR_OK) {
         TRACE_ERROR("template_update_attribute failed\n");
         free(temp_attr);
         goto done;
@@ -1019,37 +1174,37 @@ CK_RV token_specific_dh_pkcs_key_pair_gen(STDLL_TokData_t *tokdata,
     value_bits_attr->pValue =
         (CK_BYTE *) value_bits_attr + sizeof(CK_ATTRIBUTE);
     *(CK_ULONG *) value_bits_attr->pValue = 8 * temp_bn_len;
-    rc = template_update_attribute(priv_tmpl, value_bits_attr);
-    if (rc != CKR_OK) {
+    rv = template_update_attribute(priv_tmpl, value_bits_attr);
+    if (rv != CKR_OK) {
         TRACE_ERROR("template_update_attribute failed\n");
         free(temp_attr);
         goto done;
     }
 
     // Add prime and base to the private key template
-    rc = build_attribute(CKA_PRIME,
+    rv = build_attribute(CKA_PRIME,
                          (unsigned char *) prime_attr->pValue,
                          prime_attr->ulValueLen, &temp_attr);  // in bytes
-    if (rc != CKR_OK) {
+    if (rv != CKR_OK) {
         TRACE_DEVEL("build_attribute failed\n");
         goto done;
     }
-    rc = template_update_attribute(priv_tmpl, temp_attr);
-    if (rc != CKR_OK) {
+    rv = template_update_attribute(priv_tmpl, temp_attr);
+    if (rv != CKR_OK) {
         TRACE_ERROR("template_update_attribute failed\n");
         free(temp_attr);
         goto done;
     }
 
-    rc = build_attribute(CKA_BASE,
+    rv = build_attribute(CKA_BASE,
                          (unsigned char *) base_attr->pValue,
                          base_attr->ulValueLen, &temp_attr);     // in bytes
-    if (rc != CKR_OK) {
+    if (rv != CKR_OK) {
         TRACE_DEVEL("build_attribute failed\n");
         goto done;
     }
-    rc = template_update_attribute(priv_tmpl, temp_attr);
-    if (rc != CKR_OK) {
+    rv = template_update_attribute(priv_tmpl, temp_attr);
+    if (rv != CKR_OK) {
         TRACE_ERROR("template_update_attribute failed\n");
         free(temp_attr);
         goto done;
@@ -1088,14 +1243,16 @@ CK_RV token_specific_get_mechanism_list(STDLL_TokData_t *tokdata,
                                         CK_MECHANISM_TYPE_PTR pMechanismList,
                                         CK_ULONG_PTR pulCount)
 {
-    return ock_generic_get_mechanism_list(tokdata, pMechanismList, pulCount, NULL);
+    return ock_generic_get_mechanism_list(tokdata, pMechanismList, pulCount,
+                                          &token_specific_filter_mechanism);
 }
 
 CK_RV token_specific_get_mechanism_info(STDLL_TokData_t *tokdata,
                                         CK_MECHANISM_TYPE type,
                                         CK_MECHANISM_INFO_PTR pInfo)
 {
-    return ock_generic_get_mechanism_info(tokdata, type, pInfo, NULL);
+    return ock_generic_get_mechanism_info(tokdata, type, pInfo,
+                                          &token_specific_filter_mechanism);
 }
 
 CK_RV token_specific_sha_init(STDLL_TokData_t *tokdata, DIGEST_CONTEXT *ctx,
@@ -1124,6 +1281,20 @@ CK_RV token_specific_sha_final(STDLL_TokData_t *tokdata, DIGEST_CONTEXT *ctx,
     return openssl_specific_sha_final(tokdata, ctx, out_data, out_data_len);
 }
 
+CK_RV token_specific_shake_key_derive(STDLL_TokData_t *tokdata, SESSION *sess,
+                                      CK_MECHANISM *mech,
+                                      OBJECT *base_key_obj,
+                                      CK_KEY_TYPE base_key_type,
+                                      OBJECT *derived_key_obj,
+                                      CK_KEY_TYPE derived_key_type,
+                                      CK_ULONG derived_key_len)
+{
+    return openssl_specific_shake_key_derive(tokdata, sess, mech,
+                                             base_key_obj, base_key_type,
+                                             derived_key_obj, derived_key_type,
+                                             derived_key_len);
+}
+
 CK_RV token_specific_hmac_sign_init(STDLL_TokData_t *tokdata, SESSION *sess,
                                     CK_MECHANISM *mech, CK_OBJECT_HANDLE Hkey)
 {
@@ -1291,6 +1462,117 @@ CK_RV token_specific_ecdh_pkcs_derive(STDLL_TokData_t *tokdata,
 
 #endif
 
+#if OPENSSL_VERSION_PREREQ(3, 0)
+CK_RV import_ibm_dilithium_key(STDLL_TokData_t *tokdata, OBJECT *obj)
+{
+    struct soft_private_data *soft_private = tokdata->private_data;
+    const struct pqc_oid *oid = NULL;
+    CK_ATTRIBUTE *attr = NULL;
+    CK_OBJECT_CLASS class;
+    CK_BYTE *data = NULL;
+    CK_ULONG data_len;
+    CK_RV rc;
+
+    if (soft_private->oqs_provider == NULL) {
+        TRACE_ERROR("The oqsprovider is not loaded\n");
+        return CKR_MECHANISM_INVALID;
+    }
+
+    rc = template_attribute_get_ulong(obj->template, CKA_CLASS, &class);
+    if (rc != CKR_OK)
+        return CKR_OK;
+
+    /* A clear IBM Dilithium key must either have a CKA_VALUE containing
+     * the SPKI or PKCS#8 encoded private key, or must have a keyform/mode
+     * value and the individual attributes
+     */
+    if (template_attribute_find(obj->template, CKA_VALUE, &attr) == TRUE &&
+        attr->ulValueLen > 0 && attr->pValue != NULL) {
+        switch (class) {
+        case CKO_PRIVATE_KEY:
+            /* Private key in PKCS#8 form is present in CKA_VALUE */
+            rc = ibm_dilithium_priv_unwrap(obj->template, attr->pValue,
+                                           attr->ulValueLen, FALSE);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("Failed to decode private key from "
+                            "CKA_VALUE.\n");
+                return rc;
+            }
+            break;
+        case CKO_PUBLIC_KEY:
+            /* Public key in SPKI form is present in CKA_VALUE */
+            rc = ibm_dilithium_priv_unwrap_get_data(obj->template,
+                                                    attr->pValue,
+                                                    attr->ulValueLen,
+                                                    FALSE);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("Failed to decode public key from "
+                            "CKA_VALUE.\n");
+                return rc;
+            }
+            break;
+        default:
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+    } else {
+        /* Add CKA_VALUE withPKCS#8 or SPKI */
+        switch (class) {
+         case CKO_PRIVATE_KEY:
+             rc = ibm_dilithium_priv_wrap_get_data(obj->template,
+                                                   FALSE, &data, &data_len);
+             if (rc != CKR_OK) {
+                 TRACE_ERROR("Failed to encode private key.\n");
+                 return rc;
+             }
+             break;
+         case CKO_PUBLIC_KEY:
+             rc = ibm_dilithium_publ_get_spki(obj->template,
+                                              FALSE, &data, &data_len);
+             if (rc != CKR_OK) {
+                 TRACE_ERROR("Failed to encode public key.\n");
+                 return rc;
+             }
+             break;
+         default:
+             return CKR_TEMPLATE_INCONSISTENT;
+        }
+
+        rc = build_attribute(CKA_VALUE, data, data_len, &attr);
+        OPENSSL_cleanse(data, data_len);
+        free(data);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("build_attribute CKA_VALUE failed\n");
+            return rc;
+        }
+        rc = template_update_attribute(obj->template, attr);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("template_update_attribute CKA_VALUE failed.\n");
+            free(attr);
+            return rc;
+        }
+    }
+
+    oid = ibm_pqc_get_keyform_mode(obj->template, CKM_IBM_DILITHIUM);
+    if (oid == NULL) {
+        TRACE_ERROR("%s Failed to determine dilithium OID\n", __func__);
+        return CKR_TEMPLATE_INCOMPLETE;
+    }
+
+    if (openssl_get_pqc_oid_name(oid) == NULL) {
+        TRACE_ERROR("Dilithium key form is not supported by oqsprovider\n");
+        return CKR_KEY_SIZE_RANGE;
+    }
+
+    rc = ibm_pqc_add_keyform_mode(obj->template, oid, CKM_IBM_DILITHIUM);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("ibm_pqc_add_keyform_mode failed\n");
+        return rc;
+    }
+
+    return CKR_OK;
+}
+#endif
+
 CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
                                 OBJECT * obj)
 {
@@ -1301,7 +1583,9 @@ CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
 #endif
     CK_RV rc;
 
+#if !OPENSSL_VERSION_PREREQ(3, 0)
     UNUSED(tokdata);
+#endif
     UNUSED(sess);
 
     rc = template_attribute_get_ulong(obj->template, CKA_KEY_TYPE, &keytype);
@@ -1333,8 +1617,161 @@ CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
         }
         return CKR_OK;
 
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    case CKK_IBM_DILITHIUM:
+        return import_ibm_dilithium_key(tokdata, obj);
+#endif
+
     default:
         return CKR_OK;
     }
 }
 
+CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
+                                              CK_OBJECT_CLASS class,
+                                              CK_ULONG mode, TEMPLATE *tmpl)
+{
+    CK_KEY_TYPE keytype;
+    EVP_PKEY *pkey = NULL;
+    CK_RV rc;
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    struct soft_private_data *soft_private = tokdata->private_data;
+    const struct pqc_oid *oid = NULL;
+    const char *alg_name;
+#else
+
+    UNUSED(tokdata);
+    UNUSED(class);
+#endif
+
+    if (mode != MODE_UNWRAPPED)
+        return CKR_OK;
+
+    rc = template_attribute_get_ulong(tmpl, CKA_KEY_TYPE, &keytype);
+    if (rc != CKR_OK)
+        return CKR_OK;
+
+    switch (keytype) {
+    case CKK_DES:
+    case CKK_DES3:
+    case CKK_AES:
+    case CKK_AES_XTS:
+    case CKK_GENERIC_SECRET:
+#if !(NODH)
+    case CKK_DH:
+#endif
+#if !(NODSA)
+    case CKK_DSA:
+#endif
+    case CKK_RSA:
+        return CKR_OK;
+
+#ifndef NO_EC
+    case CKK_EC:
+        /* Check if OpenSSL supports the curve */
+        rc = openssl_make_ec_key_from_template(tmpl, &pkey);
+        if (pkey != NULL)
+                EVP_PKEY_free(pkey);
+        return rc;
+#endif
+
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    case CKK_IBM_DILITHIUM:
+        if (soft_private->oqs_provider == NULL) {
+            TRACE_ERROR("The oqsprovider is not loaded\n");
+            return CKR_MECHANISM_INVALID;
+        }
+
+        oid = ibm_pqc_get_keyform_mode(tmpl, CKM_IBM_DILITHIUM);
+        if (oid == NULL) {
+            TRACE_ERROR("%s Failed to determine dilithium OID\n", __func__);
+            return CKR_TEMPLATE_INCOMPLETE;
+        }
+
+        alg_name = openssl_get_pqc_oid_name(oid);
+        if (alg_name == NULL) {
+            TRACE_ERROR("Dilithium key form is not supported by oqsprovider\n");
+            return CKR_KEY_SIZE_RANGE;
+        }
+
+        /* Check if the oqsprovider supports the variant */
+        rc = openssl_make_ibm_dilithium_key_from_template(tmpl, oid,
+                                                          class ==
+                                                                CKO_PRIVATE_KEY,
+                                                           alg_name, &pkey);
+        if (pkey != NULL)
+                EVP_PKEY_free(pkey);
+
+        return rc;
+#endif
+
+    default:
+        return CKR_KEY_TYPE_INCONSISTENT;
+    }
+}
+
+#if OPENSSL_VERSION_PREREQ(3, 0)
+
+CK_RV token_specific_ibm_dilithium_generate_keypair(STDLL_TokData_t *tokdata,
+                                                    const struct pqc_oid *oid,
+                                                    TEMPLATE *publ_tmpl,
+                                                    TEMPLATE *priv_tmpl)
+{
+    struct soft_private_data *soft_private = tokdata->private_data;
+
+    if (soft_private->oqs_provider == NULL) {
+        TRACE_ERROR("The oqsprovider is not loaded\n");
+        return CKR_MECHANISM_INVALID;
+    }
+
+    return openssl_specific_ibm_dilithium_generate_keypair(tokdata, oid,
+                                                           publ_tmpl,
+                                                           priv_tmpl);
+}
+
+CK_RV token_specific_ibm_dilithium_sign(STDLL_TokData_t *tokdata,
+                                        SESSION *sess,
+                                        CK_BBOOL length_only,
+                                        const struct pqc_oid *oid,
+                                        CK_BYTE *in_data,
+                                        CK_ULONG in_data_len,
+                                        CK_BYTE *signature,
+                                        CK_ULONG *signature_len,
+                                        OBJECT *key_obj)
+{
+    struct soft_private_data *soft_private = tokdata->private_data;
+
+    if (soft_private->oqs_provider == NULL) {
+        TRACE_ERROR("The oqsprovider is not loaded\n");
+        return CKR_MECHANISM_INVALID;
+    }
+
+    return openssl_specific_ibm_dilithium_sign(tokdata, sess, length_only, oid,
+                                               in_data, in_data_len,
+                                               signature, signature_len,
+                                               key_obj);
+}
+
+CK_RV token_specific_ibm_dilithium_verify(STDLL_TokData_t *tokdata,
+                                          SESSION *sess,
+                                          const struct pqc_oid *oid,
+                                          CK_BYTE *in_data,
+                                          CK_ULONG in_data_len,
+                                          CK_BYTE *signature,
+                                          CK_ULONG signature_len,
+                                          OBJECT *key_obj)
+{
+    struct soft_private_data *soft_private = tokdata->private_data;
+
+    if (soft_private->oqs_provider == NULL) {
+        TRACE_ERROR("The oqsprovider is not loaded\n");
+        return CKR_MECHANISM_INVALID;
+    }
+
+    return openssl_specific_ibm_dilithium_verify(tokdata, sess, oid,
+                                                 in_data, in_data_len,
+                                                 signature, signature_len,
+                                                 key_obj);
+}
+
+#endif
