@@ -23,8 +23,9 @@
 #include <strings.h>
 #include <pwd.h>
 #include <grp.h>
-#include <err.h>
 #include <openssl/crypto.h>
+
+#include "platform.h"
 #include "slotmgr.h"
 #include "pkcsconf_msg.h"
 #include "p11util.h"
@@ -33,6 +34,10 @@
 #include "uri.h"
 #include "pin_prompt.h"
 
+#if defined(_AIX)
+    const char *__progname = "pkcsconf";
+#endif
+
 #define LEEDS_DEFAULT_PIN "87654321"
 #define PIN_SIZE 80
 #define BACK_SPACE 8
@@ -307,8 +312,10 @@ done:
     free(SlotList);
     if (FunctionPtr)
         FunctionPtr->C_Finalize(NULL);
+#ifndef WITH_SANITIZER
     if (dllPtr)
         dlclose(dllPtr);
+#endif
 
     return rv;
 }
@@ -975,7 +982,7 @@ CK_RV init(void)
      * error */
     /* The host machine should have the right library in the
      * LD_LIBRARY_PATH */
-    dllPtr = dlopen("libopencryptoki.so", RTLD_NOW);
+    dllPtr = dlopen(OCK_API_LIBNAME, DYNLIB_LDFLAGS);
     if (!dllPtr) {
         warnx("Error loading PKCS#11 library");
         warnx("dlopen error: %s", dlerror());
@@ -991,6 +998,10 @@ CK_RV init(void)
     }
 
     symPtr(&FunctionPtr);
+    if (!FunctionPtr) {
+        warnx("Error getting function list, C_GetFunctionList returned NULL");
+        return -1;
+    }
 
     /* If we get here we know the slot manager is running and we can use PKCS11
      * calls, so we will execute the PKCS11 Initilize command. */
