@@ -15,8 +15,6 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/prctl.h>
-#include <sys/capability.h>
 #include <grp.h>
 #include <pwd.h>
 #include <string.h>
@@ -32,7 +30,19 @@
 #define MD5_HASH_SIZE 16
 
 #define DEF_MANUFID "IBM"
-#define DEF_SLOTDESC    "Linux"
+
+#if defined(_AIX)
+    #define DEF_SLOTDESC    "AIX"
+    #include <sys/types.h>
+    #include <sys/priv.h>
+    #include <sys/procfs.h>
+    #include <userpriv.h>
+    #include <fcntl.h>
+#else
+    #include <sys/prctl.h>
+    #include <sys/capability.h>
+    #define DEF_SLOTDESC    "Linux"
+#endif
 
 typedef char md5_hash_entry[MD5_HASH_SIZE];
 md5_hash_entry tokname_hash_table[NUMBER_SLOTS_MANAGED];
@@ -118,6 +128,32 @@ int compute_hash(int hash_type, int buf_size, char *buf, char *digest)
     return 0;
 }
 
+#if defined(_AIX)
+/*
+ * Naming is hard! What's called capabilities on Linux is called privileges
+ * on AIX. To make matters worse, user privileges are also called privileges.
+ */
+int drop_capabilities(void)
+{
+    privg_t priv_set;
+
+    if (getppriv(-1, PRIV_MAXIMUM, priv_set, sizeof(priv_set)) == -1) {
+        fprintf(stderr,
+            "Failed to get process privileges: %s\n", strerror(errno));
+        return -1;
+    }
+
+    priv_clrall(priv_set);
+    if (setppriv(-1, priv_set, priv_set, priv_set, priv_set) == -1) {
+        fprintf(stderr,
+            "Failed to set process privileges: %s\n", strerror(errno));
+        return -1;
+    }
+
+    DbgLog(DL0, "All privileges sucessfully dropped.\n");
+    return 0;
+}
+#else
 /*
  * Drop any effective and permitted capabilities (if any).
  * This function is called after initialization, but before becoming a daemon.
@@ -128,7 +164,7 @@ int drop_capabilities(void)
 
     caps = cap_get_proc();
     if (caps == NULL) {
-        fprintf(stderr, "Failed to get process's capabilities: %s\n",
+        fprintf(stderr, "Failed to get process capabilities: %s\n",
                 strerror(errno));
         return -1;
     }
@@ -153,6 +189,7 @@ int drop_capabilities(void)
 
     return 0;
 }
+#endif
 
 /*
  * This function is called only when running as root user.
@@ -165,14 +202,41 @@ int drop_capabilities(void)
  */
 void drop_privileges(struct passwd *pwd)
 {
-    char program[PATH_MAX + 1] = { 0 };
+    char program[PATH_MAX + 1] = { 0, };
     char* args[] = { program, NULL };
 
+#if defined(_AIX)
+    struct psinfo ps;
+    char procpath[PATH_MAX + 1] = { NULL, };
+    int psfd;
+
+    snprintf(procpath, PATH_MAX, "/proc/%lld/psinfo", getpid());
+    psfd = open(procpath, O_RDONLY);
+    if (psfd == -1) {
+        fprintf(stderr,
+                "Failed to open procfs to read cmdname: %s\n", strerror(errno));
+        exit(1);
+    }
+
+    if (read(psfd, &ps, sizeof(ps)) == -1) {
+        fprintf(stderr, "Failed to populate psinfo: %s\n", strerror(errno));
+        exit(1);
+    }
+    close(psfd);
+    /*
+     * Copies only argv[0], as maintained by the kernel. This is because
+     * parameters are guaranteed to be separated by NULL, which is where strcpy
+     * stops copying.
+     */
+    strncpy(program, ps.pr_psargs, PATH_MAX);
+
+#else
     if (readlink("/proc/self/exe", program, PATH_MAX) == -1) {
         fprintf(stderr, "Failed to get executable file name: %s\n",
                 strerror(errno));
         exit(1);
     }
+#endif
 
     if (initgroups(pwd->pw_name, pwd->pw_gid) != 0) {
         fprintf(stderr, "Failed to set the group access list: %s\n",
@@ -218,7 +282,7 @@ void run_sanity_checks(void)
     struct group *grp;
     struct stat sbuf;
     struct dircheckinfo_s dircheck[] = {
-        { "/run/opencryptoki", S_IRWXU | S_IXGRP, 1 },
+        { RUN_DIR, S_IRWXU | S_IXGRP, 1 },
         { LOCKDIR_PATH, S_IRWXU | S_IRWXG, 0 },
         { OCK_LOGDIR, S_IRWXU | S_IRWXG, 0 },
         { CONFIG_PATH, S_IRWXU | S_IRWXG, 0 },
@@ -294,14 +358,16 @@ void run_sanity_checks(void)
     if (uid == 0)
         drop_privileges(pwd);
 
+/* AIX setppriv handles this already */
+#if !defined(_AIX)
     /* Do not allow execve() to grant additional privileges */
     if (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0) != 0) {
         fprintf(stderr, "Failed to set NO_NEW_PRIVS flag: %s\n",
                 strerror(errno));
         exit(1);
     }
-
-    if (chdir("/run/opencryptoki") != 0) {
+#endif
+    if (chdir(RUN_DIR) != 0) {
         fprintf(stderr, "Failed to set current directory: %s\n",
                 strerror(errno));
         exit(1);
@@ -320,21 +386,161 @@ int is_duplicate(md5_hash_entry hash, md5_hash_entry *hash_table)
     return 0;
 }
 
+int check_token_group(struct group *tok_grp)
+{
+    struct group grp_buf, *pkcs11_grp = NULL;
+    struct passwd *pwd;
+    int i, k, err, found, rc = 0;
+    long buf_size;
+    char *buff = NULL;
+
+    /* No further check if token group is 'pkcs11' */
+    if (strcmp(tok_grp->gr_name, PKCS_GROUP) == 0)
+        return 0;
+
+    /*
+     * Must use getgrnam_r() here, because caller is using getgrnam() which
+     * returns a pointer to a static area that would be reused/overwritten by
+     * subsequent calls to getgrnam().
+     */
+    buf_size = sysconf(_SC_GETGR_R_SIZE_MAX);
+    if (buf_size <= 0) {
+        err = errno;
+        fprintf(stderr, "sysconf(_SC_GETGR_R_SIZE_MAX) failed [errno=%s].\n",
+                strerror(err));
+        return err;
+    }
+
+retry:
+    buff = calloc(1, buf_size);
+    if (buff == NULL) {
+        fprintf(stderr, "Failed to allocate a buffer of %ld bytes.\n",
+                buf_size);
+        return ENOMEM;
+    }
+
+    errno = 0;
+    if (getgrnam_r(PKCS_GROUP, &grp_buf, buff, buf_size, &pkcs11_grp) != 0) {
+        err = (errno != 0 ? errno : ENOENT);
+        if (err == ERANGE && buf_size < 64 * 1024) {
+            free(buff);
+            buf_size *= 2;
+            goto retry;
+        }
+
+        fprintf(stderr, "Group '%s' does not exist [errno=%s].\n", PKCS_GROUP,
+                strerror(err));
+       rc = err;
+       goto done;
+    }
+
+    /* Check that all group members are also a member of the 'pkcs11' group */
+    for (i = 0; tok_grp->gr_mem[i] != NULL; i++) {
+        /* Check if user's primary group is the 'pkcs11' group */
+        errno = 0;
+        pwd = getpwnam(tok_grp->gr_mem[i]);
+        err = (errno != 0 ? errno : ENOENT);
+        if (pwd == NULL) {
+            fprintf(stderr, "USer '%s' does not exist [errno=%s].\n",
+                    tok_grp->gr_mem[i], strerror(err));
+            rc = EINVAL;
+            /* Continue to display all missing users */
+            continue;
+        }
+
+        if (pwd->pw_gid != pkcs11_grp->gr_gid) {
+            /* Check the users secondary groups */
+            for (k = 0, found = 0; pkcs11_grp->gr_mem[k] != NULL; k++) {
+                if (strcmp(tok_grp->gr_mem[i], pkcs11_grp->gr_mem[k]) == 0) {
+                    found = 1;
+                    break;
+                }
+            }
+
+            if (!found) {
+                fprintf(stderr, "User '%s' is member of the token group '%s', "
+                        "but is not a member of the '%s' group.\n",
+                        tok_grp->gr_mem[i], tok_grp->gr_name, PKCS_GROUP);
+                rc = EINVAL;
+                /* Continue to display all missing users */
+                continue;
+            }
+        }
+    }
+
+done:
+    free(buff);
+    return rc;
+}
+
 int chk_create_tokdir(Slot_Info_t_64 *psinfo)
 {
     struct stat sbuf;
     char tokendir[PATH_MAX];
     struct group *grp;
     gid_t grpid;
-    int uid, rc;
+    int uid, rc, err;
     mode_t proc_umask;
     char *tokdir = psinfo->tokname;
+    char *tokgroup = psinfo->usergroup;
     char token_md5_hash[MD5_HASH_SIZE];
 
-    /* skip if no dedicated token directory is required */
-    if (!tokdir || strlen(tokdir) == 0)
+    if (psinfo->present == FALSE)
         return 0;
 
+    proc_umask = umask(0);
+
+    if (strlen(tokgroup) == 0)
+        tokgroup = PKCS_GROUP;
+
+    /* get token group id */
+    uid = (int) geteuid();
+    errno = 0;
+    grp = getgrnam(tokgroup);
+    err = (errno != 0 ? errno : ENOENT);
+    if (!grp) {
+        fprintf(stderr, "Token group '%s' does not exist [errno=%s].\n",
+                tokgroup, strerror(err));
+        return err;
+    } else {
+        grpid = grp->gr_gid;
+    }
+
+    rc = check_token_group(grp);
+    if (rc)
+        return rc;
+
+    /*
+     * Skip if no dedicated token directory is required. If no 'tokname' is
+     * specified, the token directory name is not known, thus we can not check
+     * or create it.
+     */
+    if (!tokdir || strlen(tokdir) == 0) {
+        /*
+         * Build the md5 hash from the dll name prefixed with 'dll:' to
+         * check for duplicate tokens with no 'tokname'.
+         */
+        snprintf(tokendir, sizeof(tokendir), "dll:%s", psinfo->dll_location);
+        rc = compute_md5(tokendir, strlen(tokendir), token_md5_hash);
+        if (rc) {
+            fprintf(stderr, "Error calculating MD5 of token name!\n");
+            return -1;
+        }
+
+        /* check for duplicate token names */
+        if (is_duplicate(token_md5_hash, tokname_hash_table)) {
+            fprintf(stderr, "Duplicate token in slot %llu!\n",
+                    psinfo->slot_number);
+            return -1;
+        }
+
+        /* add entry into hash table */
+        memcpy(tokname_hash_table[psinfo->slot_number], token_md5_hash,
+               MD5_HASH_SIZE);
+
+        return 0;
+    }
+
     /* Check if the path length fits in the max path length
        (include 2 * / and 0)
      */
@@ -343,19 +549,6 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
         return -1;
     }
 
-    proc_umask = umask(0);
-
-    /* get 'PKCS11' group id */
-    uid = (int) geteuid();
-    grp = getgrnam(PKCS_GROUP);
-    if (!grp) {
-        fprintf(stderr, "%s group does not exist [errno=%d].\n", PKCS_GROUP,
-                errno);
-        return errno;
-    } else {
-        grpid = grp->gr_gid;
-    }
-
     /* calculate md5 hash from token name */
     rc = compute_md5(tokdir, strlen(tokdir), token_md5_hash);
     if (rc) {
@@ -376,13 +569,24 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
     /* sprintf checked above */
     sprintf(tokendir, "%s/%s", CONFIG_PATH, tokdir);
     rc = stat(tokendir, &sbuf);
-    if (rc != 0 && errno == ENOENT) {
+    err = errno;
+    if (rc != 0 && err == ENOENT) {
+        if (strcmp(tokgroup, PKCS_GROUP) != 0) {
+            fprintf(stderr,
+                    "Can not create token directory '%s' with a token group "
+                    "other than '%s'. You must create the token directory "
+                    "using the pkcstok_admin tool with the proper owner "
+                    "group.\n", tokendir, PKCS_GROUP);
+            umask(proc_umask);
+            return EACCES;
+        }
+
         /* directory does not exist, create it */
         rc = mkdir(tokendir, S_IRWXU | S_IRWXG);
         if (rc != 0) {
             fprintf(stderr,
-                    "Creating directory '%s' failed [errno=%d].\n",
-                    tokendir, errno);
+                    "Creating directory '%s' failed [errno=%s].\n",
+                    tokendir, strerror(errno));
             umask(proc_umask);
             return rc;
         }
@@ -390,25 +594,48 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
         rc = chown(tokendir, uid, grpid);
         if (rc != 0) {
             fprintf(stderr,
-                    "Could not set %s group permission [errno=%d].\n",
-                    PKCS_GROUP, errno);
+                    "Could not set '%s' group permission [errno=%s].\n",
+                    PKCS_GROUP, strerror(errno));
             umask(proc_umask);
             return rc;
         }
 
+    } else if (rc != 0) {
+        fprintf(stderr,
+                "Could not stat directory '%s' [errno=%s].\n", tokendir,
+                strerror(err));
+        umask(proc_umask);
+        return err;
+    } else if (sbuf.st_gid != grpid) {
+        fprintf(stderr,
+                "Directory '%s' is not owned by token group '%s'.\n",
+                tokendir, tokgroup);
+        umask(proc_umask);
+        return EACCES;
+    }
+
+    /*
+     * Can not check or create TOK_OBJ directory inside the token directory
+     * if the token group is different than 'pkcs11', because the 'pkcsslotd'
+     * user does not have permissions to access such a token directory.
+     */
+    if (strcmp(tokgroup, PKCS_GROUP) != 0) {
+        umask(proc_umask);
+        return 0;
     }
 
     /* Create TOK_OBJ directory */
     /* sprintf checked above */
     sprintf(tokendir, "%s/%s/%s", CONFIG_PATH, tokdir, OBJ_DIR);
     rc = stat(tokendir, &sbuf);
-    if (rc != 0 && errno == ENOENT) {
+    err = errno;
+    if (rc != 0 && err == ENOENT) {
         /* directory does not exist, create it */
         rc = mkdir(tokendir, S_IRWXU | S_IRWXG);
         if (rc != 0) {
             fprintf(stderr,
-                    "Creating directory '%s' failed [errno=%d].\n",
-                    tokendir, errno);
+                    "Creating directory '%s' failed [errno=%s].\n",
+                    tokendir, strerror(errno));
             umask(proc_umask);
             return rc;
         }
@@ -416,11 +643,23 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
         rc = chown(tokendir, uid, grpid);
         if (rc != 0) {
             fprintf(stderr,
-                    "Could not set %s group permission [errno=%d].\n",
-                    PKCS_GROUP, errno);
+                    "Could not set '%s' group permission [errno=%s].\n",
+                    PKCS_GROUP, strerror(errno));
             umask(proc_umask);
             return rc;
         }
+    } else if (rc != 0) {
+        fprintf(stderr,
+                "Could not stat directory '%s' [errno=%s].\n", tokendir,
+                strerror(err));
+        umask(proc_umask);
+        return err;
+    } else if (sbuf.st_gid != grpid) {
+        fprintf(stderr,
+                "Directory '%s' is not owned by token group '%s'.\n",
+                tokendir, tokgroup);
+        umask(proc_umask);
+        return EACCES;
     }
     umask(proc_umask);
     return 0;
@@ -432,15 +671,15 @@ static int create_pid_file(pid_t pid)
 
     pidfile = fopen(PID_FILE_PATH, "w");
     if (!pidfile) {
-        fprintf(stderr, "Could not create pid file '%s' [errno=%d].\n",
-                PID_FILE_PATH, errno);
+        fprintf(stderr, "Could not create pid file '%s' [errno=%s].\n",
+                PID_FILE_PATH, strerror(errno));
         return -1;
     }
 
     fprintf(pidfile, "%d\n", (int) pid);
     fflush(pidfile);
     fclose(pidfile);
-    InfoLog("PID File created");
+    InfoLog("PID file (" PID_FILE_PATH ") created");
 
     return 0;
 }
@@ -470,18 +709,30 @@ static int config_parse_slot(const char *config_file,
     struct ConfigBaseNode *c;
     int i, slot_no;
     unsigned int vers;
-    char *str;
+#if defined(_AIX)
+    char libname[1025] = {0, };
+    char toktype[5] = {0, };
+    char *substr = NULL;
+    int idx = 0;
+#endif
+    char *str = NULL;
 
     slot_no = slot->idx;
     DbgLog(DL3, "Slot: %d\n", slot_no);
 
     if (slot_no >= NUMBER_SLOTS_MANAGED) {
         ErrLog("Error parsing config file '%s': Slot number %d unsupported! "
-               "Slot number has to be less than %d!", config_file, slot_no,
+               "Slot number has to be less than %d!\n", config_file, slot_no,
                NUMBER_SLOTS_MANAGED);
         return 1;
     }
 
+    if (sinfo[slot_no].present) {
+        ErrLog("Error parsing config file '%s': Duplicate slot number %d\n",
+               config_file, slot_no);
+        return 1;
+    }
+
     memset(&sinfo[slot_no].slot_number, 0, sizeof(sinfo[slot_no].slot_number));
     sinfo[slot_no].slot_number = slot_no;
 
@@ -491,7 +742,27 @@ static int config_parse_slot(const char *config_file,
 
         if (strcmp(c->key, "stdll") == 0 &&
             (str = confignode_getstr(c)) != NULL) {
+#if defined(_AIX)
+           /*
+            * AIX bundles libraries as ranlib archives. The objects therefore
+            * need to be loaded from within the library. Instead of modifying
+            * the parser to handle the new string format, it's easier to just
+            * build the string ourselves - this only happens at daemon init,
+            * so performance isn't a concern here.
+            */
+            substr = strrchr(str, '_');
+            if (substr == NULL) /* no '_', libname is malformed */
+                return 1;
 
+            ++substr;
+            while (*substr != '.' && idx < sizeof(toktype) - 1) {
+                toktype[idx++] = (char)*substr;
+                ++substr;
+            }
+            snprintf(libname, sizeof(libname),
+                    "libpkcs11_%s.a(libpkcs11_%s.so.0)", toktype, toktype);
+            str = (char*)libname;
+#endif
             if (do_str((char *)&sinfo[slot_no].dll_location,
                        sizeof(sinfo[slot_no].dll_location),
                        config_file, c->key, str, 0))
@@ -559,6 +830,16 @@ static int config_parse_slot(const char *config_file,
             confignode_getversion(c, &sinfo[slot_no].version) == 0)
             continue;
 
+        if (strcmp(c->key, "usergroup") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+            if (do_str((char *)&sinfo[slot_no].usergroup,
+                       sizeof(sinfo[slot_no].usergroup),
+                       config_file, c->key, str, 0))
+                return 1;
+
+            continue;
+        }
+
         ErrLog("Error parsing config file '%s': unexpected token '%s' "
                "at line %d: \n", config_file, c->key, c->line);
         return 1;
@@ -722,6 +1003,53 @@ done:
     return ret;
 }
 
+static int teardown(int rc)
+{
+    term_socket_server();
+    DestroyMutexes();
+    DetachFromSharedMemory();
+    DestroySharedMemory();
+    return rc;
+}
+
+static int setup_sock_data(void)
+{
+    if (!init_socket_data(&socketData)) {
+        DestroyMutexes();
+        DetachFromSharedMemory();
+        DestroySharedMemory();
+        return 6;
+    }
+
+    if (event_support_disabled)
+        socketData.flags |= FLAG_EVENT_SUPPORT_DISABLED;
+
+    /* Create customized token directories */
+    psinfo = &socketData.slot_info[0];
+    for (int i = 0; i < NUMBER_SLOTS_MANAGED; i++, psinfo++) {
+        if (chk_create_tokdir(psinfo)) {
+            DestroyMutexes();
+            DetachFromSharedMemory();
+            DestroySharedMemory();
+            return EACCES;
+        }
+    }
+    /* setup completed successfully */
+    return 0;
+}
+
+static int setup_sock_server(void)
+{
+    if (!init_socket_server(event_support_disabled)) {
+        DestroyMutexes();
+        DetachFromSharedMemory();
+        DestroySharedMemory();
+        return 5;
+    }
+    /* setup completed successfully */
+    return 0;
+}
+
 /*****************************************
  *  main() -
  *      You know what main does.
@@ -732,7 +1060,7 @@ done:
 
 int main(int argc, char *argv[], char *envp[])
 {
-    int ret, i;
+    int ret;
 
     /**********************************/
     /* Read in command-line arguments */
@@ -807,30 +1135,13 @@ int main(int argc, char *argv[], char *envp[])
     if (!XProcUnLock())
         return 4;
 
-    if (!init_socket_server(event_support_disabled)) {
-        DestroyMutexes();
-        DetachFromSharedMemory();
-        DestroySharedMemory();
-        return 5;
-    }
+    if ((ret = setup_sock_data()) != 0)
+        return ret;
 
-    if (!init_socket_data(&socketData)) {
-        term_socket_server();
-        DestroyMutexes();
-        DetachFromSharedMemory();
-        DestroySharedMemory();
-        return 6;
-    }
-    if (event_support_disabled)
-        socketData.flags |= FLAG_EVENT_SUPPORT_DISABLED;
-
-    /* Create customized token directories */
-    psinfo = &socketData.slot_info[0];
-    for (i = 0; i < NUMBER_SLOTS_MANAGED; i++, psinfo++) {
-        ret = chk_create_tokdir(psinfo);
-        if (ret)
-            return EACCES;
-    }
+#if !defined(_AIX)
+    if ((ret = setup_sock_server()) != 0)
+        return ret;
+#endif
 
     if (drop_capabilities() != 0)
         return 7;
@@ -841,7 +1152,9 @@ int main(int argc, char *argv[], char *envp[])
     if (Daemon) {
         pid_t pid;
         if ((pid = fork()) < 0) {
+#if !defined(_AIX)
             term_socket_server();
+#endif
             DestroyMutexes();
             DetachFromSharedMemory();
             DestroySharedMemory();
@@ -871,6 +1184,22 @@ int main(int argc, char *argv[], char *envp[])
 #endif
     }
 
+#if defined(_AIX)
+    /*
+     * Set up the socket server /after/ the fork on AIX. This is because
+     * AIX closes the pollset fd before the fork, with no way of continuing
+     * to use it in the child. Therefore, pollset needs to be initialised
+     * ONLY in the client process to actually be useful.
+     * Quoting from the manpage:
+     * A process can call fork after calling pollset_create. The child process
+     * will already have a pollset ID per pollset, but pollset_destroy,
+     * pollset_ctl, pollset_query, and pollset_poll operations will fail with
+     * an errno value of EACCES.
+     */
+    if ((ret = setup_sock_server()) != 0)
+        return ret;
+#endif
+
     /*****************************************
      *
      * Register Signal Handlers
@@ -885,15 +1214,10 @@ int main(int argc, char *argv[], char *envp[])
      * We have to set up the signal handlers after we daemonize because
      * the daemonization process redefines our handler for (at least) SIGTERM
      */
-    if (!SetupSignalHandlers()) {
-        term_socket_server();
-        DestroyMutexes();
-        DetachFromSharedMemory();
-        DestroySharedMemory();
-        return 8;
-    }
+    if (!SetupSignalHandlers())
+        return teardown(8);
 
-    /* ultimatly we will create a couple of threads which monitor the slot db
+    /* ultimately we will create a couple of threads which monitor the slot db
      * and handle the insertion and removal of tokens from the slot.
      */
 
@@ -905,13 +1229,8 @@ int main(int argc, char *argv[], char *envp[])
 
 #if !defined(NOGARBAGE)
     /* start garbage collection thread */
-    if (!StartGCThread(shmp)) {
-        term_socket_server();
-        DestroyMutexes();
-        DetachFromSharedMemory();
-        DestroySharedMemory();
-        return 9;
-    }
+    if (!StartGCThread(shmp))
+        return teardown(9);
 #endif
 
     /*
@@ -931,7 +1250,7 @@ int main(int argc, char *argv[], char *envp[])
 
     /*************************************************************
      *
-     * Here we need to actualy go through the processes and verify that thye
+     * Here we need to actually go through the processes and verify that they
      * still exist.  If not, then they terminated with out properly calling
      * C_Finalize and therefore need to be removed from the system.
      * Look for a system routine to determine if the shared memory is held by
